package sfm.shp.gui;

import java.awt.event.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import java.io.*;
import java.util.*;

import sfm.shp.LabelData;
import sfm.shp.PixelData;
import sfm.shp.ShapeData;
import sfm.shp.gui.ImagePanel;
import sfm.shp.network.*;

import sfm.dataset.Dataset;
import sfm.dataset.Instance;
import sharkey_mod_nlp.dataset.input.ReadFile;
import sharkey_mod_nlp.dataset.loaders.*;

import sharkey_mod_nlp.fnn.ANNConfig;
import nasseutils.NasseBox;

public class GUI_FNN extends JPanel implements ActionListener, ItemListener, FilenameFilter
{
  protected Dataset dataset;
  private FNN nm;
  
  private Vector<JTextField> annCfgV;
  protected JTextArea jtA_out;
  protected JTextField tf_Epochs, tf_preEpochs, tf_inputNodes, tf_hiddenNodes, tf_outputNodes;
  protected JTextField tfTrainRatio, tfTrainSamps, tfTestSamps;
  protected JTextField tf_margin, tf_Learningrate, tf_Momentum, tf_Noise; 
  
  private JPanel attributePanel;
  
  private JProgressBar progressBar;
  protected JButton trainButt, testButt, loaderButt_sentences;
  protected JButton saveButt, loadButt, createNetworkButt;
  
  /**
   * 
   *
   */
  private static final String shapeDir = "datasets/shapes/";
  
  private String suffix;
  private String[] shapeFiles;
  private Vector<ShapeData> shapeV;
  private Vector<LabelData> labelV;
  private ImagePanel imgP;
  
  private JButton trainShpButt, trainLabelButt;
  private JButton trainShp2WordButt, trainWord2ShpButt;
  
  private final int hiddenUnits_shapeNet = 30;
  private final int hiddenUnits_labelNet = 16;
  
  private boolean shpWordMapperDataAdded;
  
  private FNN shapeNet, labelNet, shape2WordMapper, word2ShapeMapper;
  private ANNConfig cfg_shp, cfg_lbl, cfg_mapShape2, cfg_mapLabel2;
  
  
  public GUI_FNN()
  {
    annCfgV   = new Vector<JTextField>();
    
    cfg_shp       = new ANNConfig();
    cfg_lbl       = new ANNConfig();
    cfg_mapShape2 = new ANNConfig();
    cfg_mapLabel2 = new ANNConfig();
    
    //////////////////////////////////////////////////////////////////////////////////////
    ///// attributes panel - enable diff reduction (transform to a small set of integers)
    attributePanel = new JPanel();
    JPanel jpWest = new JPanel(new BorderLayout());    
    jpWest.add(attributePanel, BorderLayout.EAST);
    
    //  ------- checkbox
    /*
    JCheckBox cb_force = new JCheckBox("Run until successful", false); 
    cb_force.addItemListener(this);
    cb_force.setName("forcerun");
    */
    
    JCheckBox cb_updating = new JCheckBox("Update weights per epoch", false); 
    cb_updating.addItemListener(this);
    cb_updating.setName("update_pr_epoch");
    
    JCheckBox cb_prestructure = new JCheckBox("Enable prestructuring", true); 
    cb_prestructure.addItemListener(this);
    cb_prestructure.setName("prestructuring");

    JPanel cb_updatingP = new JPanel(new GridLayout(1,1));
    cb_updatingP.add(cb_prestructure);
    cb_updatingP.add(cb_updating);
    
    //------- checkbox
    
    createNetworkButt = new JButton("Create Network!");
    createNetworkButt.setActionCommand("makenetwork");
    createNetworkButt.addActionListener(this);
    
    trainButt = new JButton("Train");
    trainButt.setActionCommand("start_training");
    trainButt.addActionListener(this);
    trainButt.setEnabled(false);
    
    testButt = new JButton("Test");
    testButt.setActionCommand("start_testing");
    testButt.addActionListener(this);
    testButt.setEnabled(false);

    ////////////////////////////////////////
    /// create gui for the dataset-loaders
    JPanel loaderPanelX = new JPanel(new GridLayout(2, 1, 4, 2));
    
    loaderButt_sentences = new JButton("Load \"Sentences\" dataset");
    loaderButt_sentences.setActionCommand("load_sentences");
    loaderButt_sentences.addActionListener(this);
    
    loaderPanelX.add(loaderButt_sentences);
    
    JPanel loaderP = new JPanel(new FlowLayout(FlowLayout.LEFT));
    loaderP.add(loaderPanelX);
    
    JPanel buttP = new JPanel();
    buttP.add(createNetworkButt);
    buttP.add(trainButt);
    buttP.add(testButt);
    
    JPanel trainButtP = new JPanel(new BorderLayout());
    JLabel dummyL3 = new JLabel("-"); dummyL3.setForeground(trainButtP.getBackground());
    
    // trainButtP.add(dummyL3, BorderLayout.NORTH);
    trainButtP.add(loaderP, BorderLayout.NORTH);
    trainButtP.add(buttP, BorderLayout.WEST);
    
    tf_Epochs       = new JTextField("100");
    tf_preEpochs    = new JTextField("4000");
    tf_margin       = new JTextField("0.1");
    tf_Learningrate = new JTextField("0.25");
    tf_Momentum     = new JTextField("0.65");
    tf_Noise        = new JTextField("0.00");
    tf_Noise.setEditable(false);
    
    tf_inputNodes   = new JTextField("");
    tf_hiddenNodes  = new JTextField("");
    tf_outputNodes  = new JTextField("");
    
    // tf_inputNodes.setEditable(false);
    // tf_outputNodes.setEditable(false);
    tf_margin.setToolTipText("The strength of classification: its value is the lowest accepted output for each pattern...");
    
    progressBar = new JProgressBar(0,0);
    
    tf_Learningrate.setBorder(new BevelBorder(BevelBorder.LOWERED));
    tf_margin.setBorder(new BevelBorder(BevelBorder.LOWERED));
    tf_Momentum.setBorder(new BevelBorder(BevelBorder.LOWERED));
    tf_Noise.setBorder(new BevelBorder(BevelBorder.LOWERED));
    tf_Epochs.setBorder(new BevelBorder(BevelBorder.LOWERED));
    tf_preEpochs.setBorder(new BevelBorder(BevelBorder.LOWERED));
    
    JPanel cfgPanel = new JPanel(new GridLayout(13,2));
    cfgPanel.add(new JLabel("Margin"));			      cfgPanel.add(tf_margin);
    cfgPanel.add(new JLabel("Epochs"));			      cfgPanel.add(tf_Epochs);
    cfgPanel.add(new JLabel("Prestructuring"));   cfgPanel.add(tf_preEpochs);
    
    cfgPanel.add(new JLabel("Learning rate"));	  cfgPanel.add(tf_Learningrate);
    cfgPanel.add(new JLabel("Momentum"));			    cfgPanel.add(tf_Momentum);
    cfgPanel.add(new JLabel("Noise"));			      cfgPanel.add(tf_Noise);    
    cfgPanel.add(new JLabel(""));			            cfgPanel.add(new JLabel(""));
    cfgPanel.add(new JSeparator());			          cfgPanel.add(new JSeparator());
    
    cfgPanel.add(new JLabel("Input nodes   "));	   cfgPanel.add(tf_inputNodes);
    cfgPanel.add(new JLabel("Hidden nodes  "));	   cfgPanel.add(tf_hiddenNodes);
    cfgPanel.add(new JLabel("Output nodes  "));	   cfgPanel.add(tf_outputNodes);
    
    cfgPanel.add(new JLabel(""));			            cfgPanel.add(new JLabel(""));
    cfgPanel.add(new JLabel("Progress  "));       cfgPanel.add(progressBar);
    
    NasseBox nb = new NasseBox(null);
    nb.add(cfgPanel);

    JPanel wrapP = new JPanel(new BorderLayout(30,10));
    wrapP.add(cb_updatingP, BorderLayout.NORTH);
    wrapP.add(nb, BorderLayout.CENTER);    
    wrapP.add(trainButtP, BorderLayout.SOUTH);    
    // wrapP.add(loaderP, BorderLayout.EAST);
    
    /////////////////////////////////////
    ///// the output tray....
    jtA_out = new JTextArea(10,75);
    jtA_out.setBorder(new BevelBorder(BevelBorder.LOWERED));
    JScrollPane jsp = new JScrollPane(jtA_out, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
    
    JPanel outputP = new JPanel(new BorderLayout(2,10));
    outputP.add(new JSeparator(), BorderLayout.NORTH);
    outputP.add(new JLabel("Actual classifications & MSE for dataset"), BorderLayout.WEST);
    outputP.add(jsp, BorderLayout.SOUTH);
    
    JPanel wrapCfg = new JPanel(new FlowLayout(FlowLayout.LEFT));
    wrapCfg.add(wrapP);
    
    JPanel bothP = new JPanel(new BorderLayout(10,16));
    bothP.add(jpWest, BorderLayout.WEST);
    bothP.add(wrapCfg, BorderLayout.EAST);
    bothP.add(outputP, BorderLayout.SOUTH);
    
    this.setLayout(new FlowLayout(FlowLayout.LEFT));
    this.add(bothP);
    
    saveButt = new JButton("Save Network");
    saveButt.setActionCommand("save");
    saveButt.addActionListener(this);
    
    loadButt = new JButton("Load Network");
    loadButt.setActionCommand("getfromfile");
    loadButt.addActionListener(this);
    
    this.add(saveButt);
    this.add(loadButt);
    
    annCfgV.addElement(tf_margin);
    annCfgV.addElement(tf_Epochs);
    annCfgV.addElement(tf_preEpochs);
    annCfgV.addElement(tf_Learningrate);
    annCfgV.addElement(tf_Momentum);
    annCfgV.addElement(tf_Noise);
    annCfgV.addElement(tf_inputNodes);
    annCfgV.addElement(tf_hiddenNodes);
    annCfgV.addElement(tf_outputNodes);
    
    tf_margin.setName("margin");
    tf_Epochs.setName("epochs");
    tf_preEpochs.setName("pre_epochs");
    tf_Learningrate.setName("learningrate");
    tf_Momentum.setName("momentum");
    tf_Noise.setName("noise");    
    tf_inputNodes.setName("inputnodes");
    tf_hiddenNodes.setName("hiddennodes");
    tf_outputNodes.setName("outputnodes");
    
    
      extraGUI();
  }
//-----------------------------------------------------------------//
  private JPanel extraGUI()
  {
    JPanel groundingPanel = new JPanel();
    
    suffix     = "BMP";
    shapeV     = new Vector<ShapeData>();
    labelV     = new Vector<LabelData>();
    shapeFiles = ReadFile.getFilesInDir(shapeDir, this);

    shpWordMapperDataAdded = false;
    
    loadLabels();
    addDatasetToLabelNet();
    
    loadShapes();    
    addDatasetToShapeNet();
  
    setupMapperNetWorks();
    
    JComboBox jcb = new JComboBox(shapeV);
    jcb.addItemListener(this);
    
    imgP = new ImagePanel(shapeV, 64, 64);
    imgP.setLayout(null);
    imgP.setBackground(Color.GRAY);
    imgP.setBorder(new BevelBorder(BevelBorder.LOWERED));
    
    trainShpButt = new JButton("Learn Shapes...");
    trainShpButt.addActionListener(this);
    trainShpButt.setActionCommand("trainShapes");

    trainLabelButt = new JButton("Learn Labels...");
    trainLabelButt.addActionListener(this);
    trainLabelButt.setActionCommand("trainLabels");
    
    trainShp2WordButt = new JButton("Map Shapes 2 Labels...");
    trainShp2WordButt.addActionListener(this);
    trainShp2WordButt.setActionCommand("trainMapper_shp2Word");
    
    trainWord2ShpButt = new JButton("Map Labels 2 Shapes...");
    trainWord2ShpButt.addActionListener(this);
    trainWord2ShpButt.setActionCommand("trainMapper_word2Shp");
    
    
    JPanel jp  = new JPanel(new BorderLayout());    
    JPanel jp1 = new JPanel();
    jp1.add(jcb);
    jp1.add(imgP);
    
    JPanel jp2 = new JPanel(new GridLayout(4, 1));
    jp2.add(trainShpButt);
    jp2.add(trainLabelButt);
    jp2.add(trainShp2WordButt);
    jp2.add(trainWord2ShpButt);
    
    jp.add(jp1, BorderLayout.NORTH);
    jp.add(jp2, BorderLayout.WEST);
    
    groundingPanel.add(jp);
    
    return groundingPanel;
  }
  
//---------------------------------------------------------------------//
  private ShapeData getShapeByName(String name)
  {
    ShapeData shp;
    
    for (int i=0; i < shapeV.size(); i++)
    {
      shp = shapeV.elementAt(i);
      // System.out.println("shapeName : " + shp.getShapeName() + " :: " + name);
      if (shp.getShapeName().equalsIgnoreCase(name)) return shp;
    }
    return null;
  }
//---------------------------------------------------------------------//
  private LabelData getLabelByName(String name)
  {
    LabelData label;
    
    for (int i=0; i < labelV.size(); i++)
    {
      label = labelV.elementAt(i);
      if (label.getLabel().equalsIgnoreCase(name)) return label;
    }
    return null;
  }
//---------------------------------------------------------------------//
  private void addMapperDatasets()
  {
    double[] pat1, pat2;
    ShapeData shp;
    LabelData ld;

    Dataset ds1 = new Dataset("Shapes-2-Labels", shapeV.size());
    Dataset ds2 = new Dataset("Labels-2-Shapes", shapeV.size());
      
    // and labelNet activations as trainingTargets
    for (int i=0; i < shapeV.size(); i++)
    {
      shp = shapeV.elementAt(i);
      ld  = getLabelByName(shp.getShapeName());
        
      shapeNet.runThisInputPattern(shp.getPattern(true));
      pat1 = shapeNet.getLayerActivation(shapeNet.getHiddenLayer());
        
      labelNet.runThisInputPattern(ld.getPattern());
      pat2 = labelNet.getLayerActivation(labelNet.getHiddenLayer());
        
      ds1.addInstance(new Instance("", pat1, i, pat2));
      ds2.addInstance(new Instance("", pat2, i, pat1));
    }
    
    shape2WordMapper.setDataset(ds1);
    word2ShapeMapper.setDataset(ds2);
  }
//---------------------------------------------------------------------//
  private void setupMapperNetWorks()
  {
    // create the AnnConfigs... 
    
    // shape2WordMapper = new FNN();
    // word2ShapeMapper = new FNN();
    
  }
//---------------------------------------------------------------------//  
  private void addDatasetToShapeNet()
  {
    // use training a scheme that classifies shapes by classID
    
    shapeNet = new FNN(ShapeData.getGlobalMaxPatternLength(2), hiddenUnits_shapeNet, shapeV.size());
    
    ShapeData shp;
    Instance inst;    
    Dataset ds = new Dataset("Shapes", shapeV.size());
    
    for (int i=0; i < shapeV.size(); i++)
    {
      shp  = shapeV.elementAt(i);
      inst = new Instance(shp.getShapeName(), shp.getPattern(true), shp.getShapeID(), shp.getTrainingTarget());
      
      ds.addInstance(inst);  
    }
    shapeNet.setDataset(ds);
  }
//---------------------------------------------------------------------//
  private void addDatasetToLabelNet()
  {
    labelNet = new NativeNetWrapper(LabelData.getExpandedPatternLength(), hiddenUnits_labelNet, labelV.size());
    
    LabelData labelD;
    Instance inst;    
    Dataset ds = new Dataset("Labels", labelV.size());
    
    for (int i=0; i < labelV.size(); i++)
    {
      labelD = labelV.elementAt(i);
      inst   = new Instance(labelD.getLabel(), labelD.getPattern(), labelD.getLabelID(), labelD.getTrainingTarget());
     
      ds.addInstance(inst);
    }
    labelNet.setDataset(ds);
  }
//---------------------------------------------------------------------//
  private void loadLabels()
  {
    final double posValue = 0.5;
    
    LabelData label1 = new LabelData("TRIANGLE");
    label1.setPOS(LabelData.POS_NOUN, posValue);
    
    LabelData label2 = new LabelData("RECTANGLE");
    label2.setPOS(LabelData.POS_NOUN, posValue);
    
    LabelData label3 = new LabelData("CIRCLE");
    label3.setPOS(LabelData.POS_NOUN, posValue);
    
    LabelData label4 = new LabelData("ABOVE");
    label4.setPOS(LabelData.POS_PREP, posValue);
    
    LabelData label5 = new LabelData("BELOW");
    label5.setPOS(LabelData.POS_PREP, posValue);
    
    LabelData label6 = new LabelData("LEFT");
    label6.setPOS(LabelData.POS_PREP, posValue);
    
    LabelData label7 = new LabelData("RIGHT");
    label7.setPOS(LabelData.POS_PREP, posValue);
    
    LabelData label8 = new LabelData("ON");
    label8.setPOS(LabelData.POS_PREP, posValue);

    labelV.addElement(label1);
    labelV.addElement(label2);
    labelV.addElement(label3);
    labelV.addElement(label4);
    labelV.addElement(label5);
    labelV.addElement(label6);
    labelV.addElement(label7);
    labelV.addElement(label8);
  }
//---------------------------------------------------------------------//
  private void loadShapes()
  {
    String sN;
    
    for (int i=0; i < shapeFiles.length; i++)
    {
      sN = shapeFiles[i].toUpperCase();
      if (sN.indexOf('.') != -1) sN = sN.substring(0, sN.indexOf('.'));
      
      ShapeData shp = ReadFile.readShapeData(shapeDir + shapeFiles[i], sN);
      shapeV.addElement(shp);
    }
  }
//---------------------------------------------------------------------//
  public boolean accept(File dir, String name)
  {
    if (suffix.equals(""))
    {
      if (name.indexOf('.') == -1) return true;
      return false; 
    }
    
    if (name.toUpperCase().endsWith(suffix)) return true;
    return false;
  }
  
//-----------------------------------------------------------------//
  /**
   * This method may only be called AFTER a dataset has been loaded!
   */
  private JPanel createDatasetPanel()
  {
    ///////////////////////////////////////////////////////
    JLabel tfdatasetName     = new JLabel(dataset.getName());
    JTextField tfdatasetSize = new JTextField("" + dataset.getNumberOfInstances());
    JTextField tfdatasetCls  = new JTextField("" + dataset.getNumberOfClasses());    
    
    tfTrainSamps  = new JTextField();    
    tfTestSamps   = new JTextField();    
    tfTrainRatio  = new JTextField("0.75", 4);
    
    tfTrainSamps.setEditable(false);
    tfTestSamps.setEditable(false);
    
    tfdatasetSize.setEditable(false);
    tfdatasetCls.setEditable(false);
    
    tfdatasetName.setFont(new Font("Courier New", Font.BOLD, 12));    
    JLabel dName = new JLabel("Dataset = ");
    dName.setFont(new Font("Courier New", Font.BOLD, 12));
    
    JPanel infoPanel = new JPanel(new GridLayout(8,2));
    
    infoPanel.add(dName);
    infoPanel.add(tfdatasetName);

    JLabel invis0 = new JLabel("invisible");
    invis0.setForeground(infoPanel.getBackground());    
    JLabel invis00 = new JLabel("invisible");
    invis00.setForeground(infoPanel.getBackground());
    infoPanel.add(invis0);    
    infoPanel.add(invis00);
    
    infoPanel.add(new JLabel("Training Ratio"));
    infoPanel.add(tfTrainRatio);
    
    infoPanel.add(new JLabel("Instances"));
    infoPanel.add(tfdatasetSize);
    
    infoPanel.add(new JLabel("Classes"));
    infoPanel.add(tfdatasetCls);

    JLabel invis1 = new JLabel("invisible");
    invis1.setForeground(infoPanel.getBackground());    
    JLabel invis11 = new JLabel("invisible");
    invis11.setForeground(infoPanel.getBackground());
    infoPanel.add(invis1);    
    infoPanel.add(invis11);
    
    infoPanel.add(new JLabel("Training size"));
    infoPanel.add(tfTrainSamps);
    
    infoPanel.add(new JLabel("Testing size"));
    infoPanel.add(tfTestSamps);
    
    // c = tfTrainRatio.getBackground();
    
    JPanel infoWrap = new JPanel(new FlowLayout(FlowLayout.LEFT));
    infoWrap.add(infoPanel);
    //------------------------------------------
    
    JPanel jspView = new JPanel(new BorderLayout());
    jspView.setBorder(new BevelBorder(BevelBorder.LOWERED));    
    jspView.add(infoWrap, BorderLayout.NORTH);
    
    return jspView;
  }
//---------------------------------------------------------------------//
  public void clearOutput()
  {
    jtA_out.setText("");
  }
//---------------------------------------------------------------------//
  public void print(String s)
  {
    jtA_out.append(s);
  } 
//---------------------------------------------------------------------//
  public void showProgress(int progressValue)
  {
    progressBar.setValue(progressValue);
  }
//---------------------------------------------------------------------//  
  public void notify_finished()
  {
    trainButt.setEnabled(true);
    progressBar.setValue(0);
  }
//---------------------------------------------------------------------//  
  public void actionPerformed(ActionEvent e)
  {
    String cmd = e.getActionCommand();
    
    if (parseInputfields())
    {
      if (cmd.equalsIgnoreCase("trainShapes"))
      {
        shapeNet.startTraining(0.25, 0.65, 500, true);
      }
      else if (cmd.equalsIgnoreCase("trainLabels"))
      {
        labelNet.startTraining(0.45, 0.90, 5000, false);
      }
      else if (cmd.equalsIgnoreCase("trainMapper_shp2Word"))
      {
        if (!shpWordMapperDataAdded)
        {
          addMapperDatasets();
          shpWordMapperDataAdded = true;
        }
        shape2WordMapper.startTraining(0.15, 0.75, 400, true);
      }
      else if (cmd.equalsIgnoreCase("trainMapper_word2Shp"))
      {
        if (!shpWordMapperDataAdded)
        {
          addMapperDatasets();
          shpWordMapperDataAdded = true;
        }
        word2ShapeMapper.startTraining(0.15, 0.75, 100, false);
      }
    }
  }
//---------------------------------------------------------------------//
  
//---------------------------------------------------------------------//
  
//---------------------------------------------------------------------//  
  public void itemStateChanged(ItemEvent e)
  {
    if (e.getSource() instanceof JCheckBox)
    {
      JCheckBox cb = (JCheckBox) e.getSource();
      
      if (cb.getName().equalsIgnoreCase("forcerun"))
      {
        if (e.getStateChange() == ItemEvent.DESELECTED)
        {
          annConfig.setRunUntilSuccessful(false);
          tf_Epochs.setEnabled(true);
        }
        else
        {
          annConfig.setRunUntilSuccessful(true);
          tf_Epochs.setEnabled(false);
        }
      }
      else if (cb.getName().equalsIgnoreCase("update_pr_epoch"))
      {
        if (e.getStateChange() == ItemEvent.DESELECTED)
        {
          annConfig.setUpdateWeightsPerEpoch(false);          
        }
        else
        {
          annConfig.setUpdateWeightsPerEpoch(true);
        }
      }
      else if (cb.getName().equalsIgnoreCase("prestructuring"))
      {
        if (e.getStateChange() == ItemEvent.DESELECTED)
        {
          annConfig.setUsePrestructuring(false);          
        }
        else
        {
          annConfig.setUsePrestructuring(true);
        }
      }
    }
    else 
    {
      if (e.getStateChange() == ItemEvent.SELECTED)
      {
        ShapeData shp = (ShapeData) e.getItem();
        imgP.setSelectedShape(shp.getShapeID());
        imgP.repaint();
        
        // System.out.println("Pixels in shape : " + shp.getNumberOfPixelsInShape());
      }
    }
  }  
//---------------------------------------------------------------------//
  /**
   * Parses all textfields and sets the corresponding values in the config
   * template for our ANN; ANNConfig.
   * All flags from checkboxes are updated on-the-fly... 
   */
  protected boolean parseInputfields()
  {
    JTextField tf = null;
    
    try
    {
      for (int i=0; i < annCfgV.size(); i++)
      {
        tf = annCfgV.elementAt(i);
        if (!tf.isEnabled()) continue;
        
        double value = Double.parseDouble(tf.getText().trim());
        
        boolean ok = annConfig.setValueFor(tf.getName(), value);        
        tf.setBackground(Color.WHITE);
        
        // some buggy stuff; tried to set the value for a named target that were NOT found, stop.
        if (!ok) return false;
      }
    }
    catch (NumberFormatException e)
    {
      if (tf != null) tf.setBackground(Color.PINK);
      return false;
    }
    
    progressBar.setMinimum(0);
    progressBar.setMaximum(annConfig.getEpochs());
    
    ////////////////////////////////
    double trainRatio = 0.75;
    
    try
    {
      trainRatio = Double.parseDouble(tfTrainRatio.getText().trim());
    }
    catch (NumberFormatException ex) 
    {
      tfTrainRatio.setText("<invalid>");
      // tfTrainRatio.setBackground(Color.PINK);
    }
    
    annConfig.setTrainingRatio(trainRatio);
    
    return true;
  }
//---------------------------------------------------------------------//  
}

class ImagePanel extends JPanel
{
  private Vector<ShapeData> shpV;
  private Dimension myDim;
  private int shpPtr;
  
  public ImagePanel(Vector<ShapeData> shp, int w, int h)
  {
    shpV   = shp;
    myDim  = new Dimension(w, h);
    shpPtr = 0;
  }
//---------------------------------------------------------------------//
  public void setSelectedShape(int shpNo)
  {
    shpPtr = shpNo;
  }
//---------------------------------------------------------------------//
  public void paint(Graphics g)
  {
    super.paint(g);
    
    g.translate(16,16);
    g.setColor(Color.WHITE);
    
    Vector<PixelData> v = shpV.elementAt(shpPtr).getEdgePixels();    
    int x, y;
    
    for (int i=0; i < v.size(); i++)
    {
      x = v.elementAt(i).getX();
      y = v.elementAt(i).getY();
      g.drawLine(x, y, x, y);
    }
  }
//---------------------------------------------------------------------//  
  public void update(Graphics g)
  {
    // super.update(g);
  }
//---------------------------------------------------------------------//
  public Dimension getSize()          { return myDim; }
  public Dimension getMinimumSize()   { return myDim; }
  public Dimension getMaximumSize()   { return myDim; }
  public Dimension getPreferredSize() { return myDim; }
//---------------------------------------------------------------------//
}