package sfm.shp.network;

import java.io.Serializable;

public final class Neuron implements Serializable
{
  Link[] forwardConnections, backwardConnections;
  private double netInput, calculatedOutputFunc;
  private boolean inputChanged;
  public double tmpError;
//------------------------------------------------//
  public Neuron(int backwardLinks, int forwardLinks)
  { 
    netInput = 0.0;
    
    if (backwardLinks > 0) backwardConnections = new Link[backwardLinks];    
    if (forwardLinks >  0) forwardConnections  = new Link[forwardLinks];
    
    // weightChanged = true;
    inputChanged  = true;
  }
//------------------------------------------------//
  public double nodeFunction()
  {
    if (inputChanged)
    {
      calculatedOutputFunc = (1 / (1 + (Math.exp(-netInput))));
      inputChanged = false;
    }
    return calculatedOutputFunc;
  }
//------------------------------------------------//
  public void addInput(double input)
  {
    inputChanged = true;
    netInput += input;
  }
//------------------------------------------------//
  public void propagateForward()
  {
    Neuron n;
    
    for (int i=0; i < forwardConnections.length; i++)
    {
      n = forwardConnections[i].getOutNeuron();
      n.addInput(forwardConnections[i].getWeight() * nodeFunction());
    }      
  }
//------------------------------------------------//
  /**
   * Does not run the input through the nodeFunction;
   */
  public void propagateForwardRaw()
  {
    Neuron n;
    
    for (int i=0; i < forwardConnections.length; i++)
    {
      n = forwardConnections[i].getOutNeuron();
      n.addInput(forwardConnections[i].getWeight() * netInput);
    }      
  }
//-------------------------------------------------------//
  public final double getBackwardPrevChangeAt(int place)
  {
    return backwardConnections[place].prevChange;
  }
  
//-------------------------------------------------------//
  public final void realizeBackwardLinks()
  {
    for (int wI=0; wI < backwardConnections.length; wI++)
    {
      backwardConnections[wI].realizeChanges();
    }
  }
//------------------------------------------------//
  /**
   * SETS the net input to 'value';
   * Used for resetting the total input of
   * a node to make it ready for the next input-pattern.
   */
  public void setInput(double value)
  {
    inputChanged  = true;
    this.netInput = value;
  }
//------------------------------------------------//
  public double getInput()
  {
    return this.netInput;
  }
//------------------------------------------------//
  public void setForwardLink(int linkNo, Link link)
  {
    forwardConnections[linkNo] = link;
  }
//------------------------------------------------//
  public void setBackwardLink(int linkNo, Link link)
  {
    backwardConnections[linkNo] = link;
  }
//------------------------------------------------//
  public Link[] XXX_getForwardLinks()
  {
    return forwardConnections;
  }
//------------------------------------------------//
  public Link[] XXX_getBackwardLinks()
  {
    return backwardConnections;
  }
//------------------------------------------------//
}