package sharkey_mod_nlp.dataset;

import java.io.Serializable;

public final class Attribute implements Serializable
{
	/**
	 * The 'integerValue' is supposed to be equvivalent to
	 * the real-value 'attributeValue' - but is usually
	 * also reduced/simplified to a smaller range, to
	 * facilitate usage in DTC and NBC.
	 * The raw 'attributeValue' are not simplified this way,
	 * in current implementation. (ANN's don't need cheating). 
	 */
  private String attributeName, attributeValueStr;
  private double attributeValue;
  private int attributeNo;
  private long integerValue;
  private boolean hasNumericValue;
  
  public Attribute(int id, String name, double value)
  {
    this.attributeName     = name;
    this.attributeValue    = value;
    this.attributeValueStr = null;    
    this.attributeNo       = id;
    hasNumericValue        = true;
  }
//---------------------------------------------------------------------//
  public Attribute(int id, String name, String strValue)
  {
    this.attributeName     = name;
    this.attributeNo       = id;
    this.hasNumericValue   = false;
  	this.attributeValueStr = strValue;
  	this.integerValue      = 0;    // convertStr2Integer(strValue);
  	this.attributeValue    = 0.0;   // 1.0 / Math.log10(integerValue);
  }
//---------------------------------------------------------------------//
  public Attribute cloneMe()
  {
  	Attribute cAttrib = new Attribute(attributeNo, attributeName, attributeValue);
  	
  	cAttrib.attributeValueStr = this.attributeValueStr;
  	cAttrib.integerValue      = this.integerValue;
  	cAttrib.hasNumericValue   = this.hasNumericValue;
  	
  	return cAttrib;
  }
//---------------------------------------------------------------------//
  private long convertStr2Integer(String s)
  {
  	String cumStr = "";
  	byte[] strBytes = s.getBytes();
  	
  	int sVal;
  	for (int i=0; i < strBytes.length; i++)
  	{
  		try
    	{
    		sVal    = Integer.parseInt((int)strBytes[i] + "");
    		cumStr += sVal + "";
    	}
    	catch (NumberFormatException e)
    	{
    	}
  	}
  	
  	if (cumStr.length() > 10) cumStr = cumStr.substring(0, 10);  	
  	long value = Long.parseLong(cumStr);
  	
  	return (long) (value / 389.0);
  }
//---------------------------------------------------------------------//
  public boolean hasNumericValue()
  {
    return this.hasNumericValue;
  }
//---------------------------------------------------------------------//
  public int    getAttributeNo()  { return this.attributeNo; }  
  public double getValue()        { return this.attributeValue; }
  public long   getSimpleValue()  { return this.integerValue; }
  public String getValueTxt()     { return this.attributeValueStr; }
  public String getName()         { return this.attributeName; }
//---------------------------------------------------------------------//
  public void setSimpleValue(long value)
  {
    integerValue = value;
  }
//---------------------------------------------------------------------//  
  public void setRealValue(double value)
  {
    attributeValue = value;
  }
//---------------------------------------------------------------------//
}