package sharkey_mod_nlp.dataset;

import java.io.Serializable;
import java.util.Vector;

public class Lexicon implements Serializable
{
  /**
   * sequence of letters, not necessarily unambigous to a specific 
   * word in a specific sense. Ex.: 'train' will have a POS-code where
   * the bits for both NOUN and VERB are set.
   */
  private Vector<POSTag> posTable;
  private Vector<Lexeme> lexTable;
  
  public Lexicon()
  {
    posTable = new Vector<POSTag>();
    lexTable = new Vector<Lexeme>();
  }
//----------------------------------------------------------//  
  public void addEntry(String lex, int posMask, int[] posBits, int id, int patternWidth)
  {
    String bin  = Integer.toBinaryString(posMask);
    String zPad = "";
    for (int i=0; i < (patternWidth - bin.length()); i++) zPad += "0";
    
    lexTable.addElement(new Lexeme(lex, posMask, posBits, id, zPad + bin));
  }
//----------------------------------------------------------//
  public void addPOS_TAG(String[] data, int id)
  {
    posTable.addElement(new POSTag(data[0], data[1], id));
  }
//----------------------------------------------------------//
  public int getPosIDFor(String posName)
  {
    for (int i=0; i < posTable.size(); i++)
    {
      if (posTable.elementAt(i).getPOSName().equalsIgnoreCase(posName))
      {
        return posTable.elementAt(i).getPOS_ID();
      }
    }
    return -1;
  }
//----------------------------------------------------------//
  public Vector<Lexeme> getWordsInDictionary()
  {
    return this.lexTable;
  }
//----------------------------------------------------------//
  public String getRandomLexemeOfPOS(int posID)
  {
    Vector<Lexeme> words = getWordsOfPOSType(posID);
    
    int rndIdx = (int) (words.size() * Math.random());
    
    return words.elementAt(rndIdx).getLexeme();
  }
//----------------------------------------------------------//
  public Vector<Lexeme> getWordsOfPOSType(int posID)
  {
    Vector<Lexeme> posWords = new Vector<Lexeme>();
    
    Lexeme l;
    
    for (int i=0; i < lexTable.size(); i++)
    {
      l = lexTable.elementAt(i);
      
      if (l.hasPOSbitSet(posID))
      {
        posWords.addElement(l);
      }
    }
    
    return posWords;
  }
//----------------------------------------------------------//
  public Vector<POSTag> getPOSInDictionary()
  {
    return this.posTable;
  }
//----------------------------------------------------------//
  public int[] getLexemeBitmaskFor(String lexeme)
  {
    int[] res = {0,-1};
    
    for (int i=0; i < lexTable.size(); i++)
    {
      if (lexTable.elementAt(i).getLexeme().equalsIgnoreCase(lexeme))
      {
        res[0] = lexTable.elementAt(i).getPOS_bitmasx();
        res[1] = 0;
        return res;
      }
    }
    return res;
  }
//----------------------------------------------------------//
  public int getLexemeIDFor(String lexeme)
  {
    for (int i=0; i < lexTable.size(); i++)
    {
      if (lexTable.elementAt(i).getLexeme().equalsIgnoreCase(lexeme))
      {
        return lexTable.elementAt(i).getID();
      }
    }
    return -1;
  }
//----------------------------------------------------------//
  
}
//----------------------------------------------------------//








/**
 * 
 * Table for keeping all defind PART-OF-SPEECH'es (POS)
 *
 */
final class POSTag implements Serializable
{
  private String label, nameDesc;
  private int posID;
  
  public POSTag(String label, String nameDescription, int id)
  {
    this.label    = label;
    this.nameDesc = nameDescription;
    this.posID    = id;
  }
//----------------------------------------------------------//
  public String getPOSName()     { return label;    }
  public int getPOS_ID()         { return posID;    }
  public String getDescription() { return nameDesc; }
//----------------------------------------------------------//
}