package sharkey_mod_nlp.dataset.input;

import java.io.*;
import java.util.*;
import javax.imageio.*;
import java.awt.image.*;
import sfm.shp.*;

/**
 * This is a general utility-class for reading datasets from files.
 * The format of the data is assumed to be in txt-format, with one instance
 * pr. line...
 */
public abstract class ReadFile
{
//----------------------------------------------------------------------//
  public static String[] getFilesInDir(String dir, FilenameFilter filter)
  {
    File dirF      = new File(dir);    
    String[] files = null;
    
    if (filter == null)
    {
      files = dirF.list();
    }
    else
    {
      files = dirF.list(filter);
    }
   
    return files;
  }
//----------------------------------------------------------------------//
  /**
   * For reading shape-date from image-files (e.g. JPG or BMP)
   */
  public static ShapeData readShapeData(String fileName, String shapeName)
  {
    try
    {
      BufferedImage bufImg = ImageIO.read(new File(fileName));
      ShapeData shp        = new ShapeData(shapeName, bufImg);
      return shp;      
    }
    catch (FileNotFoundException e)
    {
      System.out.println("File: " + fileName + " NOT found!");
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    } 
    return null;
  }
//----------------------------------------------------------------------//  
  public static Vector<String> readDatasetFromFile(String fileName)
  {
    Vector<String> data = new Vector<String>();
    
    try
    {
      BufferedReader reader = new BufferedReader(new FileReader(fileName));
      String line           = "";

      while (line != null)
      {
        line = reader.readLine();
        
        if (line != null)
        {
          line = line.trim();
          
          // ignore commented lines....
               if (line.equalsIgnoreCase("")) continue; 
          else if (line.startsWith(";"))      continue;
          else if (line.startsWith("#"))      continue;
          else if (line.startsWith("//"))     continue;
          
          data.addElement(line.trim());
        }
      }
      
      reader.close();
      return data;      
    }
    catch (FileNotFoundException e)
    {
      System.out.println("Dataset file: " + fileName + " NOT found!");
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }    
    return null;
  }
//-------------------------------------------------------------------------//
}
