package sharkey_mod_nlp.fnn;

import java.util.Vector;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;

import sharkey_mod_nlp.dataset.Instance;
import sharkey_mod_nlp.fnn.raam.*;
import sharkey_mod_nlp.gui.*;
import sharkey_mod_nlp.gui.tree.TreeCanvas;
import sharkey_mod_nlp.gui.tree.TreeFrame;
import sharkey_mod_nlp.treemodel.*;

public class RAAM extends FNN implements Runnable, Serializable
{
  private ParseTreeData[] treeData;
  
  public RAAM(GUI_RAAM gui)
  {
    super(gui);
  }
//-----------------------------------------------------------------//
  public RAAM(GUI_RAAM gui, ANNConfig annCfg)
  {
    super(gui, annCfg);
    extractTrainingStructures();
  }
//-----------------------------------------------------------------//  
  public void saveMe(String toFile)
  {
    // saves the nodes, topology, connectivity, weights...
    super.saveMe(toFile);
    
    try
    {
      FileOutputStream fs    = new FileOutputStream(toFile + ".output");
      ObjectOutputStream out = new ObjectOutputStream(fs); 
      
      out.writeObject(topLevelPatterns);
      out.writeObject(treeData);
      
      /*
      for (int i=0; i < topLevelPatterns.size(); i++)
      {
        double[] pat = topLevelPatterns.elementAt(i);
        RAAM.printArray("raam-pats", pat, 5);
      }
      */
      
      out.flush();
      out.close();
    }
    catch (FileNotFoundException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
//-----------------------------------------------------------------//  
  public ANNConfig loadMe(String fromFile)
  {
    super.loadMe(fromFile);
    
    try
    {
      FileInputStream fs   = new FileInputStream(fromFile + ".output");
      ObjectInputStream in = new ObjectInputStream(fs); 
      
      topLevelPatterns = (Vector<Instance>) in.readObject();
      treeData         = (ParseTreeData[]) in.readObject();
      
      in.close();
      return annCfg;
    }
    catch (FileNotFoundException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    catch (ClassNotFoundException e)
    {
      e.printStackTrace();
    }
    return null;
  }
//-----------------------------------------------------------------//
  public Vector<Instance> getTopLevelPatterns()
  {
    return topLevelPatterns;
  }
//-----------------------------------------------------------------//
  public void extractTrainingStructures()
  {
    Vector<ParseNode> parseTreeV = dataset.getParseTrees();    
    treeData = new ParseTreeData[parseTreeV.size()]; 
    
    /**
     * Store data in a more accessible way to optimize for
     * speed during training.
     */
    for (int i=0; i < parseTreeV.size(); i++)
    {
      Tree<ParseNode> t     = new Tree<ParseNode>(parseTreeV.elementAt(i));      
      Vector<ParseNode>[] v = t.getNodesAsPly_Arrays();      
      treeData[i]           = new ParseTreeData(v);
    }
  }
//-----------------------------------------------------------------//
  private Vector<ParseNode[]> extractLeafNodesByGroup(Vector<ParseNode>[] parseTree)
  {
    Tree<ParseNode> tree         = new Tree<ParseNode>(parseTree[0].elementAt(0));    
    Vector<ParseNode> nodes      = tree.getAllNodes();
    Vector<ParseNode> tmpGroup   = new Vector<ParseNode>();
    
    ParseNode child, parent;
    Vector<IGenericNode> children;
    boolean pureGroup;
    
    ParseNode[] pureNodeGroup;
    Vector<ParseNode[]> nodeGroupV = new Vector<ParseNode[]>(); 

    // remove all nonLeaves and 'impure' nodeGROUPS: grps containing NON-leafnodes. 
    for (int i=0; i < nodes.size();) // no i++ inc here!
    {
      child = nodes.elementAt(i);
      
      if (!child.isLeafNode())
      {
        i++;
        continue;
      }
      
      parent   = child.getParent();
      children = parent.getChildren();

      pureGroup = true;
      tmpGroup.removeAllElements();
      
      for (int j=0; j < children.size(); j++)
      {
        child = (ParseNode) children.elementAt(j);
        
        if (!child.isLeafNode())
        {
          pureGroup = false;
        }
        
        tmpGroup.addElement(child);
        i++;
      }
      
      if (pureGroup) 
      {
        pureNodeGroup = new ParseNode[tmpGroup.size()];
        
        for (int k=0; k < tmpGroup.size(); k++) 
        {
          pureNodeGroup[k] = tmpGroup.elementAt(k);
        }
        nodeGroupV.addElement(pureNodeGroup);
      }
    }
    
    return nodeGroupV;
  }
//-----------------------------------------------------------------//
  /**
   * This method is in SERIOUS need for optimization!
   * Even though the input to this method is ONE pattern, (a structured parse of a sentence),
   * each of its attributes (single words or POS elements) must be trained separately,
   * Backpropagating for each input pair (all nodes withing a group)! 
   */
  private double runThisInput(Vector<ParseNode>[] parseTree, boolean training)
  {
    Vector<ParseNode> curPlyVector;
    ParseNode tmpNode, parent;
    int nodesInGroup;
    double[] pattern = null;
    double patternError = 0.0, lrBoost = 0.0;
    Vector<double[]> patternV = new Vector<double[]>();
    
    // start at bottom of tree (the leaf-nodes), stop before the root (ply=0).
    for (int i = (parseTree.length-1); i > 0; i--)
    {
      // this vector contains ALL nodes the this ply, i. Each nodegroup must be processed SEPARATELY!      
      curPlyVector = parseTree[i];      
      if (curPlyVector == null || curPlyVector.isEmpty()) continue;
      
      for (int j=0; j < curPlyVector.size();)
      {
        parent       = curPlyVector.elementAt(j).getParent();
        nodesInGroup = parent.getChildren().size();
        
        // Each node in this group becomes part of the inputpattern.
        // max # of nodes in a group equals max VALENCY of the parsetrees,
        // and as such the partitioning 'n' for the RAAM inputlayer...
        for (int k=0; k < nodesInGroup; k++)
        {
          tmpNode = curPlyVector.elementAt(j++);
          pattern = tmpNode.getBitpattern();
          patternV.addElement(pattern);
        }
        
        /**
         * FINALLY, we have a complete inputpattern; autoassociate it...
         */
        resetAllInputs();
        setConcInput(patternV);
        // propagateInput();
        propagateInput(patternV.size() * pattern.length);
        
        if (training)
        {
          // if (i > 2) lrBoost = 0.0;
          // else lrBoost = 0.0;
          
          patternError += backPropagate_RAAM(patternV.size() * pattern.length, 1.0);
          updateWeights();
          storeHiddenLayerActivation(parent);
        }
        patternV.removeAllElements();
      } 
    }
    
    return patternError;
  }
//-----------------------------------------------------------------//
  /**
   * Propagates values from inputnodes that has any input. This does NOT mean
   * that an input-node with total input == ZERO is ignored, but rather refers to
   * the number of partition that are used for a given input. The valency of a (sub)tree.
   * ==> Enables us to avoid propagating potentially "disturbing" activations from
   * inputnodes in inactive partitions! Does it help...? 
   */
  private void propagateInput(int nodesWithInput)
  {
    for (int i=0; i < nodesWithInput; i++)
    {
      inputLayer[i].propagateForward();
    }
    
    for (int i=0; i < hiddenLayer.length; i++) hiddenLayer[i].propagateForward();
  }
//-----------------------------------------------------------------//
  private final void resetAllInputs()
  {
    for (int i=0; i < inputLayer.length;  i++)  inputLayer[i].setInput(0.0);
    for (int i=0; i < hiddenLayer.length; i++)  hiddenLayer[i].setInput(0.0);
    for (int i=0; i < outputLayer.length; i++)  outputLayer[i].setInput(0.0);
  }
//-----------------------------------------------------------------//
  private final void storeHiddenLayerActivation(final ParseNode node)
  {
    // store the hiddenlayer activation, in parentNode
    // in next i-loop (the ply above) we use the bitpatterns stored now...
    
    double[] actPat = new double[hiddenLayer.length];
    
    for (int i=0; i < hiddenLayer.length; i++)
    {
      actPat[i] = hiddenLayer[i].nodeFunction();
      // actPat[i] = hiddenLayer[i].getInput();
    }
    
    // System.out.println("Storing hiddenActivation for " + node.getNodeLabel());
    
    node.setBitpattern(actPat);
  }
//-----------------------------------------------------------------//
  private final void setConcInput(Vector<double[]> patternV)
  {
    double[] pattern;
    int idx = 0;
    
    for (int i=0; i < patternV.size(); i++)
    {
      pattern = patternV.elementAt(i);
      
      for (int j=0; j < pattern.length; j++)
      {
        inputLayer[idx++].setInput(pattern[j]);
      }
    }
  }
//-----------------------------------------------------------------//
  private final void setHiddenLayerInput(double[] pattern)
  {
    for (int i=0; i < pattern.length; i++)
    {
      hiddenLayer[i].setInput(pattern[i]);
    }
  }
//-----------------------------------------------------------------//
  /**
   * Changed to comply with RAAM-type training (out = in)
   * 
   * No need to do error-correction on partitions not used:
   * if valency of any autoassoc operation is less than max, we can
   * skip weight correction for these outputNodes: set error for these to ZERO.
   */
  private final double backPropagate_RAAM(int lastNodeWithInput, double lrBoostMag)
  {
    double errorK = 0, errorJ = 0, patternError = 0.0;
    double momentumAdd, wChange, outputK, inputK, outputJ;
    int j, i;
    
    computeInputAndHiddenLayerNFs();
    
    // for each outputNode to hiddenNodes
    for (i=0; i < outputLayer.length; i++)
    {
      outputK = outputLayer[i].nodeFunction();
      inputK  = inputLayer[i].getInput();
      
      if (i >= lastNodeWithInput)
      {
        errorK = 0.0;
      }
      else
      {
        errorK        = outputK * (1.0 - outputK) * (inputK - outputK);
        patternError += Math.abs((inputK - outputK));
        // patternError += ((inputK - outputK) * (inputK - outputK));
      }
            
      outputLayer[i].tmpError = errorK;
      
      // for each backwardlink in outputNode[i]... update the linkWeight back to hiddenLayer-nodes
      for (j=0; j < outputLayer[i].backwardConnections.length; j++)
      {
        // wChange = ((learningRate+lrBoost) * errorK * hiddenLayer[j].nodeFunction()); // the pure weight change
        wChange = ((learningRate*lrBoostMag) * errorK * hiddenLayerNF[j]); // the pure weight change
        
        momentumAdd = (momentum * outputLayer[i].getBackwardPrevChangeAt(j));  // adding momentum-stuff! 
        outputLayer[i].backwardConnections[j].cumulateChanges(wChange + momentumAdd);
      }
    }
    
    // for each hiddenNode...to inputnodes
    for (i=0; i < hiddenLayer.length; i++)
    {
      outputJ = hiddenLayer[i].nodeFunction();
      errorJ  = (outputJ * (1.0 - outputJ)) * getSumOfWeightedErrors(hiddenLayer[i]);      
      
      for (j=0; j < hiddenLayer[i].backwardConnections.length; j++)
      {
        // wChange = ( (learningRate+lrBoost) * (errorJ * inputLayer[j].nodeFunction())); // the pure weight change
        wChange = ( (learningRate*lrBoostMag) * (errorJ * inputLayerNF[j])); // the pure weight change
        
        momentumAdd = (momentum * hiddenLayer[i].getBackwardPrevChangeAt(j));  // adding momentum-stuff!
        hiddenLayer[i].backwardConnections[j].cumulateChanges(wChange + momentumAdd);
      }
    }
    return patternError;
  }
  
//--------------------------------------------------------------------//
  /**
   * Input-param 'leafNodes' contains ALL leafNodes in ALL patterns AND where
   * any such node are part of a PURE leafNode group! They are already grouped by
   * their respective group, so each 'leafGroup' retrieved will be part of inputpattern...
   */
  private double runPreTrainInput(Vector<ParseNode[]> leafNodes, boolean storeHiddenActivation)
  {
    Vector<double[]> patternV = new Vector<double[]>();    
    int patternSize           = leafNodes.elementAt(0)[0].getBitpattern().length;
    ParseNode[] leafGroup;
    double epochError = 0.0;
    
    for (int i=0; i < leafNodes.size(); i++)
    {
      patternV.removeAllElements();
      leafGroup = leafNodes.elementAt(i);

      for (int j=0; j < leafGroup.length; j++)
      {
        patternV.addElement(leafGroup[j].getBitpattern());
      }

      resetAllInputs();
      setConcInput(patternV);
      propagateInput(patternV.size() * patternSize);
      epochError += backPropagate_RAAM(patternV.size() * patternSize, 1.0);
      updateWeights();
      
      if (storeHiddenActivation) storeHiddenLayerActivation(leafGroup[0].getParent());
    }
    
    return epochError;
  }
//--------------------------------------------------------------------//
  private void preTrain()
  {
    ///////////////////////////////////////////////////////////
    /// Start by pre-structuring weights by leafnode-autoassoc    
    Vector<ParseNode[]> leafNodes = new Vector<ParseNode[]>(); 
    
    // for each pattern - extract leafnodes for pre-training
    for (int i=0; i < treeData.length; i++)
    {
      Vector<ParseNode[]> tmpNodes = extractLeafNodesByGroup(treeData[i].getParseTreeData());
      leafNodes.addAll(tmpNodes);
    }
      
    double epochError = 0.0;
    
    double lrCopy   = learningRate;
    double momCopy  = momentum;
    
    learningRate = 0.18;
    momentum     = 0.70;
    /////////////////////////////
    /// Now train them up good
    for (int i=0; i < preStructuringEpochs; i++)
    {
      epochError = runPreTrainInput(leafNodes, i == (preStructuringEpochs-1) );
    }
    
    // restore to original values for normal training...
    learningRate = lrCopy;
    momentum     = momCopy;
    
    System.out.println("final preTrain epochError = " + epochError);
  }
//--------------------------------------------------------------------//
  /**
   * |Probably| ready for RAAM-type training now
   */
  protected void startTrainingX()
  {
    gui.clearOutput();
    rnd.setSeed(System.currentTimeMillis());

    double epochError;
    int patternNo;
    ParseTreeData inst;
    
    Vector<ParseTreeData> tmpPatterns = new Vector<ParseTreeData>();
    Vector<Double> stats              = new Vector<Double>();
    long startS                       = System.currentTimeMillis();
    
    ////////////////////////////////////////////////
    // Start pretraining to shape the weightMatrix
    if (annCfg.usePrestructuring) preTrain();
    
    
    ///////////////////////////////////////////////////////////
    for (int i=0; i < epochs; i++)
    {
      gui.showProgress(i+1);
      
      ////////////////////////////////////
      /// ADD random pattern-order! HERE:
      for (int j=0; j < treeData.length; j++) tmpPatterns.addElement(treeData[j]);
      epochError = 0;
      
      while (!tmpPatterns.isEmpty())
      {
        patternNo   = rnd.nextInt(tmpPatterns.size());
        inst        = tmpPatterns.elementAt(patternNo);
        epochError += runThisInput(inst.getParseTreeData(), true);
        
        tmpPatterns.remove(patternNo);
      }
      
      stats.addElement(new Double(epochError / treeData.length));
      // if (isFinished()) break;
    }
    /////////////////////////////////////////////////////////     
    long endS = System.currentTimeMillis();
    
    
    
    /**
     * Run one last time through all patterns, now storing the top-level
     * pattern for each class :- required for training the FNN(mapper module)
     */
    Instance instance;
    topLevelPatterns = new Vector<Instance>(); 
    for (int i=0; i < treeData.length; i++)
    {
      inst = treeData[i];
      runThisInput(inst.getParseTreeData(), false);
      
      instance = new Instance("n/a", null, inst.getSyntacticClass());
      instance.setPlainANN_data(getPatternFromLayer(hiddenLayer.length, hiddenLayer));
      topLevelPatterns.addElement(instance);
    }
    
    printResults(((endS - startS) / 1000.0), stats, treeData);
  }
//---------------------------------------------------------------------//
  
  // MAKE'em configurable parameters + include test to distinguish terminal/non-t
  private final double SLACK_TERMINAL    = 0.2;
  private final double SLACK_NONTERMINAL = 0.02;
  
  private double cumulDecodeError;
  private Vector<DebugInfo> failedDecodings;
  private ParseNode resultTree;
  
  /**
   * decodes the pattern (rootNode) encoded into the supplied
   * pattern 'parseTree'.
   * 
   * Param 'buildTree' determines whether this input should be decoded quickly (only
   * returning success/failure for the input as a whole), or actually completely reconstructing
   * a parsetree, showing all details of the decoding process and the computed parse. This is,
   * of course, slower and need not be used during TRAINING, unless we plan to use the error-
   * details to improve training/error-correction...
   */
  private double decodeThis(final Vector<ParseNode>[] parseTree)
  {
    cumulDecodeError = 0.0;
    failedDecodings  = new Vector<DebugInfo>();
    
    resultTree       = new ParseNode(null, "Root(S)");
    resultTree.setTextualValue(new String[] {"S"});
    resultTree.setPly(0);
        
    ParseNode parent = parseTree[0].elementAt(0);
    decode_knownSample(parent, parent.getBitpattern());
    
    return cumulDecodeError;
  }
//---------------------------------------------------------------------//
  
  /**
   * For use in REAL tesing!
   */
  private void decodeForReal(ParseNode referenceNode, double[] pattern, ParseNode newTree)
  {
    resetAllInputs();
    setHiddenLayerInput(pattern);
    propagateHiddenInputRAW();

    Vector<IGenericNode> children    = referenceNode.getChildren(); 
    Vector<ParseNode> clonedChildren = new Vector<ParseNode>();
    double[] referenceCurPat, actualPattern;
    ParseNode child, childClony;
    
    //----------------------------------------
    // for each information-carrying partition...
    for (int i=0; i < children.size(); i++)
    {
      child      = (ParseNode)children.elementAt(i);
      childClony = child.getPrimitiveClone();
      childClony.setParent(newTree);
      
      newTree.addChild(childClony);
      clonedChildren.addElement(childClony);

      referenceCurPat = child.getBitpattern();
      
      // match retrieved patterns from outputlayer with stored bitpatterns
      if (child.isLeafNode())
      {
        if (!slackCompareReal(referenceCurPat, i, true))
        {
          // mark this node as NOT equal to reference pattern?
          // ...
        }
      }
      else
      {
        if (!slackCompareReal(referenceCurPat, i, false))
        {
          // mark this node as NOT equal to reference pattern?
          // ...
        }
      }
      
      childClony.setDecodedBitpattern(getPatternFromOutputPartition(i, pattern.length));
    }
    //----------------------------------------
    
    // recurse at will...
    for (int i=0; i < children.size(); i++)
    {
      child      = (ParseNode)children.elementAt(i);
      childClony = clonedChildren.elementAt(i);
      
      if (!child.isLeafNode())
      {
        // actualPattern = child.getDecodedBitpattern();
        actualPattern = childClony.getDecodedBitpattern();
        decodeForReal(child, actualPattern, clonedChildren.elementAt(i));
      }
    }
  }
//---------------------------------------------------------------------//
  
  /**
   * Yeah...  for use during training!
   */
  private void decode_knownSample(ParseNode curParent, double[] pattern)
  {
    resetAllInputs();
    setHiddenLayerInput(pattern);
    propagateHiddenInputRAW();

    Vector<IGenericNode> children = curParent.getChildren(); 
    double[] curPat;
    ParseNode child;
    
    //----------------------------------------
    // for each information-carrying partition...
    for (int i=0; i < children.size(); i++)
    {
      child = (ParseNode)children.elementAt(i);

      // match retrieved patterns from outputlayer with stored bitpatterns
      if (child.isLeafNode())
      {
        DebugInfo db = new DebugInfo();
        curPat       = child.getBitpattern();
        
        if (!slackCompare(db, curPat, i, SLACK_TERMINAL))
        {
          db.nodeLabel  = child.getNodeLabel();
          db.node       = child;
          db.refOutput  = curPat;
          
          failedDecodings.addElement(db);
        }
      }
      else // it's not a leafnode, it's a non-terminal : decode it 
      {
        child.setDecodedBitpattern(getPatternFromOutputPartition(i, pattern.length));
      }
    }
    //----------------------------------------
    
    // recurse at will...
    for (int i=0; i < children.size(); i++)
    {
      child = (ParseNode)children.elementAt(i);
      
      if (!child.isLeafNode())
      {
        curPat = child.getDecodedBitpattern();
        decode_knownSample(child, curPat);
      }
    }
  }
//---------------------------------------------------------------------//
  private double[] getPatternFromOutputPartition(int partitionNo, int nBits)
  {
    double[] actualOutput = new double[nBits];
    int idx               = (partitionNo * nBits);
    
    for (int i=0; i < actualOutput.length; i++)
    {
      actualOutput[i] = outputLayer[idx++].nodeFunction();
    }
    
    return actualOutput;
  }
//---------------------------------------------------------------------//
  /**
   * for use during real testing
   */
  private boolean slackCompareReal(double[] lookupPattern, int pn, boolean computeError)
  {
    double[] actualOutput = getPatternFromOutputPartition(pn, lookupPattern.length);
    
    Vector<Integer> correctHiBits = findNHighestOuputs(ParseNode.getNumberOfBitsON(), lookupPattern);
    Vector<Integer> actualHiBits  = findNHighestOuputs(ParseNode.getNumberOfBitsON(), actualOutput);
    
    /**
     * compute the error for this sub-pattern (partition 'pn'
     */
    if (computeError)
    {
      double absDiff;
      
      for (int i=0; i < lookupPattern.length; i++)
      {
        absDiff = (lookupPattern[i] - actualOutput[i]);
        absDiff = Math.abs(absDiff);
        
        cumulDecodeError += absDiff;
      }
    }
    //-------------------------------
    
    /**
     * Checks the pattern to find the n highest bits/output-nodes, and matches this
     * against a reference pattern; if all the same n bits are highest ('on'), the
     * pattern is considered "equal" enough. We don't care about what the values 
     * are...at this point.
     */
    if (hasSameContent(correctHiBits, actualHiBits))
    {
      boolean[] correctBits = new boolean[lookupPattern.length];
      
      for (int i=0; i < correctBits.length; i++) correctBits[i] = true;
      return true;
    }
    return false;
    /**
     * returns only upon definite failure, as when the x bits who should be ON,
     * are not. If the x bits have the 'logically' correct ON/OFF characteristics,
     * a more stringent test of actual values may still fail the decoded pattern!
     */
    //------------------------------------------------------------
  }
//---------------------------------------------------------------------//
  
  /**
  * Useful 'slack' comparison of patterns where we know the general structure;
  * where a specific number of bits are to be 'ON' and 'OFF' (including their position).
  * For the present decoding application, use on TERMINALS / leaf-nodes only.
  * 
  * Obviously should not be used with intermediate patterns, like hidden-unit activations;
  * the non-terminals.
  * 
  * ::->for use during training<-::
  */
  private boolean slackCompare(DebugInfo db, double[] lookupPattern, int pn, double maxError)
  {
    boolean[] correctBits = new boolean[lookupPattern.length];
    db.actualOutput       = getPatternFromOutputPartition(pn, lookupPattern.length);
    
    Vector<Integer> correctHiBits = findNHighestOuputs(ParseNode.getNumberOfBitsON(), lookupPattern);
    Vector<Integer> actualHiBits  = findNHighestOuputs(ParseNode.getNumberOfBitsON(), db.actualOutput);

    
    /**
     * compute the error for this sub-pattern (partition 'pn'
     */
    double absDiff;
    
    for (int i=0; i < lookupPattern.length; i++)
    {
      absDiff = (lookupPattern[i] - db.actualOutput[i]);
      absDiff = Math.abs(absDiff);
      
      cumulDecodeError += absDiff;
    }      

    
    /**
     * Checks the pattern to find the n highest bits/output-nodes, and match this
     * against a reference pattern; if all the same n bits are highest ('on'), the
     * pattern is considered "equal" enough. We don't care about what the values are...
     */
    if (hasSameContent(correctHiBits, actualHiBits))
    {
      db.slackCompSuccess = true;
      
      for (int i=0; i < correctBits.length; i++) correctBits[i] = true;
      db.bitResDetails = correctBits;
      return true;
    }
    else return false;
    //------------------------------------------------------------
  }  
//---------------------------------------------------------------------//
  /**
   * Propgagates the input directly; does not run it through the nodeFunction.
   * :: Utility-method for handling input that are actually previous hiddenlayer-
   * activations, and therefore should NOT be transformed further!
   */
  private void propagateHiddenInputRAW()
  {
    for (int i=0; i < hiddenLayer.length; i++) hiddenLayer[i].propagateForwardRaw();
  }
//---------------------------------------------------------------------//
  protected void printResults(double secs, Vector stats, ParseTreeData[] patterns)
  {
    // System.out.println("RAAM.line.1.of.printResults(...)");
    
    ///////////////////////////////////////////////////////////
    ///// run through all patterns again,  with no training
    int sux = 0;
    double lastBackpropError, patternError;
    Vector<ParseNode>[] instance;

    gui.clearOutput();
    gui.print("FAILED PATTERN CLASSIFICATIONS:");
    
    for (int i=0; i < patterns.length; i++)
    {
      instance     = patterns[i].getParseTreeData();      
      patternError = decodeThis(instance);
      
      if (failedDecodings.isEmpty())
      {
        sux++;
        
        // System.out.println("Correct decodings:");
        // String nfo = instance[0].elementAt(0).getNodeLabel();
        // nfo += " (class " + instance[0].elementAt(0).getSyntacticStructureID() + ")";
        // System.out.println("\n" + nfo + " error = " + patternError);
      }
      else
      {
        String nfo = instance[0].elementAt(0).getNodeLabel();
        nfo += " (class " + instance[0].elementAt(0).getSyntacticStructureID() + ")";
        
        gui.print("\n" + nfo + " error = " + patternError);
        
        
        /*
        System.out.println("Failed decodings: " + failedDecodings.size());
        DebugInfo db;
        
        for (int z=0; z < failedDecodings.size(); z++)
        {
          db = failedDecodings.elementAt(z);
          
          System.out.print(db.nodeLabel + "  - slackCompSuccess = " + db.slackCompSuccess);
          System.out.println();
          for (int x=0; x < db.refOutput.length; x++)
          {
            System.out.println(db.refOutput[x] + " vs " + db.actualOutput[x] + " :: " + db.bitResDetails[x]);
          }
          System.out.println();
        }
        */
      }
    }
    
    gui.print("\n------------------------------------");
    
    if (stats != null) 
    {
      lastBackpropError = ((Double)stats.lastElement()).doubleValue();
      gui.print("\n\nBackprop MSE for last epoch : " + lastBackpropError);
      gui.print("\nEpochs run         : " + epochs);
      gui.print("\nTime elapsed       : " + secs + " secs");
    }    
    double pct = ((double)sux / (double)patterns.length) * 100.0;
    gui.print("\nSUCCESS-RATE : " + pct + "%");
  }
//---------------------------------------------------------------------//
  /**
   * Similarity measure: sum of differences -> 'error' for each element
   * in the pattern returned as one value. The lowest sum indicates highest similarity. 
   */
  private ParseNode XXXfindMostSimilarRootPattern(double[] pattern)
  {
    ParseNode root = null;
    double difference;
    double[] res;
    
    // each element is composed of a double[2]; elem[0] = index, elem[1] = errorValue
    Vector<double[]> indexOfParseTreeAndErrorValueV = new Vector<double[]>();
    
    for (int i=0; i < treeData.length; i++)
    {
      root = treeData[i].getRoot();
      res  = XXXfindDifferenceMagnitude(pattern, root.getBitpattern());
      
      difference = res[0];
      
      // store each difference, .. attaching classID / syntactic structure
      indexOfParseTreeAndErrorValueV.addElement(new double[] {i, difference});
      
      // RAAM.printArray("with this ", root.getBitpattern(), 4);
    }
    
    /**
     * Now find the lowest errorvalue, along with the corresponding 
     * syntactic structure (class id)
     */
    
    double minError = indexOfParseTreeAndErrorValueV.elementAt(0)[1];
    double tmpError;
    int indexOfTree = 0;
    for (int i=1; i < indexOfParseTreeAndErrorValueV.size(); i++)
    {
      tmpError = indexOfParseTreeAndErrorValueV.elementAt(i)[1];
      
      if (tmpError < minError)
      {
        tmpError     = minError;
        indexOfTree  = i;
      }
    }
    
    root = treeData[indexOfTree].getRoot();    
    return root;
  }
//---------------------------------------------------------------------//
  /**
   * finds the average difference between two 'vectors' of equal length,
   * by comparing elements (a double) at same indices.
   * 
   * return 0.0 if patterns are identical, so about so...
   */
  private double[] XXXfindDifferenceMagnitude(double[] p1, double[] p2)
  {
    if (p1.length != p2.length)
    {
      System.out.println("findDifferenceMagnitude(): unequal array lengths!");
      return null;
    }
    
    double diffSize = 0.0;
    double sumP1 = 0.0;
    double sumP2 = 0.0;
    
    for (int i=0; i < p1.length; i++)
    {
      diffSize += Math.abs(p1[i] - p2[i]);
      
      sumP1 += p1[i];
      sumP2 += p2[i];
    }
    
    return new double[] { diffSize, sumP1, sumP2 }; // / p1.length;
  }
//---------------------------------------------------------------------//
  public static boolean hasSameContent(Vector<Integer> valuesV1, Vector<Integer> valuesV2)
  {
    if (valuesV1.size() != valuesV2.size()) return false;
    
    for (int i=0; i < valuesV1.size(); i++)
    {
      for (int j=0; j < valuesV2.size(); j++)
      {
        if (valuesV1.elementAt(i) == valuesV2.elementAt(j))
        {
          valuesV2.remove(j);
          break;
        }
      }
    }
    
    /*
    System.out.println("=========================================");
    for (int i=0; i < valuesV1.size(); i++) System.out.print(valuesV1.elementAt(i) + " ");
    System.out.println();
    for (int i=0; i < valuesV1.size(); i++) System.out.print(valuesV1.elementAt(i) + " ");
    System.out.println("\n");
    */
    
    return valuesV2.isEmpty();
  }
//---------------------------------------------------------------------//
  public static double sameContentRatio(Vector<Integer> valuesV1, Vector<Integer> valuesV2)
  {
    double rt = 0.0;
    
    for (int i=0; i < valuesV1.size(); i++)
    {
      for (int j=0; j < valuesV2.size(); j++)
      {
        if (valuesV1.elementAt(i) == valuesV2.elementAt(j))
        {
          rt++;
          valuesV2.remove(j);
          break;
        }
      }
    }
    return rt / valuesV2.size();
  }
//---------------------------------------------------------------------//
  public static Vector<Integer> findNHighestOuputs(int nHighest, double[] orgPat)
  {
    Vector<Integer> nHighestV = new Vector<Integer>();
    Vector<Double>  patV      = new Vector<Double>();
    double tmpMax;
    int tmpIdx;
    boolean[] unUsedPos = new boolean[orgPat.length];
    
    for (int i=0; i < orgPat.length; i++)
    {
      patV.addElement(orgPat[i]);
      unUsedPos[i] = true;
    }
    
    while (nHighestV.size() < nHighest)
    {
      tmpMax = 0.0;
      tmpIdx = -1;
      
      for (int i=0; i < patV.size(); i++)
      {
        if ((patV.elementAt(i) > tmpMax) && (unUsedPos[i]))
        {
          tmpMax = patV.elementAt(i);
          tmpIdx = i;
        }
      }
      
      if (tmpIdx == -1) // crap!
      {
        System.out.println("findNHighestOuputs() in RAAM module did not terminate properly!");
        break; 
      }
        
      nHighestV.addElement(tmpIdx);
      unUsedPos[tmpIdx] = false;
    }
    
    return nHighestV; 
  }
//---------------------------------------------------------------------//
  /**
   * returns a parsetree...
   * NB: param 'classID' may not be known; in that case 
   * it is set to -1. Check it!
   */
  public ParseNode decodeThisPattern(double[] pattern, int classID)
  {
    // ParseNode probablyCorrectTree = findMostSimilarRootPattern(pattern);
    
    Vector<ParseNode> newTreesV = new Vector<ParseNode>();
    double tmpMinError  = 9999.0;
    int indexOfBestTree = -1;
    
    for (int i=0; i < treeData.length; i++)
    {
      cumulDecodeError = 0.0;
      resultTree       = treeData[i].getRoot().getPrimitiveClone();
      
      decodeForReal(treeData[i].getRoot(), pattern, resultTree);
      newTreesV.addElement(resultTree); // save 'resultTree' before next iteration (in which this ptr will be lost)
      
      if (cumulDecodeError < tmpMinError)
      {
        tmpMinError     = cumulDecodeError;
        indexOfBestTree = i;       
      }
    }
    
    ParseNode probablyCorrectTree = newTreesV.elementAt(indexOfBestTree); 
    
    /*
    System.out.println("raam-decode: recived patttern has classID = " + classID);
    System.out.println("--> interpreted as class = " + probablyCorrectTree.getSyntacticStructureID());
    System.out.println("cumulDecodeError = " + cumulDecodeError);
    System.out.println("------------------------------------------------------\n");
    */
    
    return probablyCorrectTree;
  }
//---------------------------------------------------------------------//  
  
  
  
  
  
  
  
//---------------------------------------------------------------------// 
  /**
   * non-critical Utility functions...
   */
  public static String formatNumber(double d, int fracDigits)
  {
    DecimalFormat decF = new DecimalFormat();
    DecimalFormatSymbols dfs = decF.getDecimalFormatSymbols();
    dfs.setDecimalSeparator('.');
    decF.setDecimalFormatSymbols(dfs);
    decF.setMaximumFractionDigits(fracDigits);
    
    return decF.format(d);
  }
//---------------------------------------------------------------------//
  /**
   * non-critical Utility function...
   */
  public static void printArray(String label, double[] arr, int fracDigits)
  {
    if (fracDigits == -1)
    {
      System.out.print(label + ": ");
      for (int i=0; i < arr.length; i++)
      {
        System.out.print(arr[i] + " ");
      }
      System.out.println("");
      return;
    }
    
    String formatted;
    String digs;
    int digCnt;
    
    System.out.print(label + ": ");
    for (int i=0; i < arr.length; i++) 
    {
      digs = "";
      
      formatted = formatNumber(arr[i], fracDigits);
      
      digCnt = formatted.substring(formatted.indexOf(".")+1, formatted.length()).length();
      if (digCnt < fracDigits)
      {
        for (int j=0; j < (fracDigits - digCnt); j++) digs += "0";
      }
      
      if (formatted.length() == 1)
      {
        digs = "";
        for (int j=0; j < fracDigits; j++) digs += "0";
        formatted += "." + digs;
      }
      else formatted += digs;
      
      if (!formatted.startsWith("-")) formatted = " " + formatted;
      
      System.out.print(formatted + " ");
    }
    System.out.println();
  }
//-----------------------------------------------------------------//
}