package sharkey_mod_nlp.fnn;

import java.util.*;
import java.io.*;

import sharkey_mod_nlp.dataset.*;
import sharkey_mod_nlp.gui.*;
import sharkey_mod_nlp.treemodel.*;

public class SRN extends FNN implements Runnable, Serializable
{
  // one element pr. syntactic class (8).
  private Vector<LegalSuccessor[]> legalSuccessorV;
  
  /**
   * When loaded from file
   */
  public SRN(GUI_ANN gui)
  {
    super(gui);
    commonInit();
  }
//-----------------------------------------------------------------//
  public SRN(GUI_ANN gui, ANNConfig annCfg)
  {
    super(gui, annCfg);
    commonInit();
    
    //  create tables of legal successors for each word in each of the 8 syntactic classes
    // ... and store in 'legalSuccessorV'.
    createSuccessorTables();
  }
//-----------------------------------------------------------------//
  private void commonInit()
  {
    realInputNodes  = (nInputNodes - nHiddenNodes);
    legalSuccessorV = new Vector<LegalSuccessor[]>();
  }
//-----------------------------------------------------------------//
  public void saveMe(String toFile) // Silje...
  {
    // saves the nodes, topology, connectivity, weights...
    super.saveMe(toFile);
    
    try
    {
      FileOutputStream fs    = new FileOutputStream(toFile + ".output");
      ObjectOutputStream out = new ObjectOutputStream(fs); 
      
      out.writeObject(patternsFromSRN);
      
      /*
      for (int i=0; i < patternsFromSRN.size(); i++)
      {
        double[] pat = patternsFromSRN.elementAt(i).getPlainNeuralNetData();
        RAAM.printArray("srn-pats", pat, 3);
      }
      */
      
      out.flush();
      out.close();
    }
    catch (FileNotFoundException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
//-----------------------------------------------------------------//  
  public ANNConfig loadMe(String fromFile)
  {
    super.loadMe(fromFile);
    
    try
    {
      FileInputStream fs   = new FileInputStream(fromFile + ".output");
      ObjectInputStream in = new ObjectInputStream(fs); 
      
      patternsFromSRN = (Vector<Instance>) in.readObject();
      
      in.close();
      return annCfg;
    }
    catch (FileNotFoundException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    catch (ClassNotFoundException e)
    {
      e.printStackTrace();
    }
    return null;
  }
//-----------------------------------------------------------------//
  /**
   * Returns the patterns that resulted from training the SRN
   * to "successfully" represent whole sentences in one pattern;
   * these patterns are taken from the hiddenlayer-outputFunctions.
   * 
   *  In current application, these will be used as the training-data
   *  for the mapper-module...the input patterns.
   */
  public Vector<Instance> getCompressedPatterns()
  {
    return patternsFromSRN;
  }
//-----------------------------------------------------------------//
  public void createSuccessorTables()
  {
    // use the 8 parsetrees to extract the valid successors for each leaf in tree,
    // then create the boolean[74] array giving which *words* may follow! 
    Vector<ParseNode> rootV     = dataset.getParseTrees();
    Vector<ParseNode> posTermV  = null;
    LegalSuccessor[] successors = null;
    boolean[] sucTable;
    
    // for each syntactic class (8)
    for (int i=0; i < rootV.size(); i++)
    {
      Tree<ParseNode> t = new Tree<ParseNode>(rootV.elementAt(i));      
      posTermV          = t.getLeafNodes();      
      successors        = new LegalSuccessor[posTermV.size()];
      
      /**
       * for each word/term in sentence (in current syntactic class)...
       * 
       * -1 because we find successors by looking at the POS of the word 
       * AFTER the current word (element #j)
       */
      for (int j=0; j < successors.length-1; j++)
      {
        sucTable      = getSuccessors(posTermV.elementAt(j), posTermV.elementAt(j+1));
        successors[j] = new LegalSuccessor(posTermV.elementAt(j).getNodeLabel(), sucTable);
      }
      
      // add successors for the last word in a sentence manually, having NO successors.
      // except maybe the PERIOD POS '.'
      sucTable = getSuccessors(null, null);
      successors[successors.length-1] = new LegalSuccessor(posTermV.lastElement().getNodeLabel(), sucTable);
      
      legalSuccessorV.addElement(successors);
    }
    
    /*
    LegalSuccessor[] testSux = legalSuccessorV.elementAt(0);
    for (int j=0; j < testSux.length; j++)
    {
      boolean[] tbl = testSux[j].getLegalSuccessorTable();
      
      System.out.println(testSux[j].getPOSLabel() + ":");
      for (int i=0; i < tbl.length; i++)
      {
        System.out.println(tbl[i] + " ");
      }
      System.out.println("\n-------------------");
    }
    */
  }
//-----------------------------------------------------------------//
  private boolean[] getSuccessors(ParseNode src, ParseNode successorPOS)
  {
    Lexicon lex         = dataset.getLexicon();
    boolean[] sucTable  = new boolean[lex.getWordsInDictionary().size()];

    //-------------------------------------------
    // not neccessary, but to be SURE every element is initialized to false
    for (int i=0; i < sucTable.length; i++) sucTable[i] = false;    
    if (successorPOS == null) return sucTable;
    //-------------------------------------------
    
    Vector<Lexeme> lexV = lex.getWordsOfPOSType(successorPOS.getPOS_ID());

    // System.out.println("Successors for " + src.getNodeLabel() + " in SYN_CLASS 0 -:- " + successorPOS.getNodeLabel());
    for (int i=0; i < lexV.size(); i++)
    {
      sucTable[lexV.elementAt(i).getID()] = true;
      // System.out.println(lexV.elementAt(i).getLexeme());
    }
    
    return sucTable;
  }
//-----------------------------------------------------------------//
  private static int instCnt = 0;
  
  public void generateSentences()
  {
    Vector<String[]> sentenceV = new Vector<String[]>(); 
    String[] aSentence;
    
    // for pre-structuring
    int shortUnambigous      = 2547;
    
    // for trainingset
    int structsPrClass_train = 180; // = 1440
    
    // for testset
    int stuctsPrClass_test   = 20;  // = 160
    
    Lexicon lex = dataset.getLexicon();
    Vector<ParseNode> sStructures = dataset.getParseTrees();
    
    ParseNode structRoot;
    instCnt = 0;
    
    // for each syntactic class... generate 'x' sentences
    for (int i=0; i < sStructures.size(); i++)
    {
      structRoot = sStructures.elementAt(i);
      
      Tree<ParseNode> t        = new Tree<ParseNode>(structRoot);
      Vector<ParseNode> leaves = t.getLeafNodes();
      
      for (int j=0; j < structsPrClass_train; j++)
      {
        // pick a word at random, for the POS-type of each leaf... creating one sentence
        createASentence(lex,  leaves, sentenceV);
      }
      
      writeToFile(sentenceV, i);
      sentenceV.removeAllElements();
    }
  }
//----------------------------------------------------//
  private void createASentence(Lexicon lex, Vector<ParseNode> leaves, Vector<String[]> sentenceV)
  {
    String[] aSentence = new String[leaves.size()];
    ParseNode leaf;
    String word;
    
    for (int j=0; j < leaves.size(); j++)
    {
      leaf         = leaves.elementAt(j);
      word         = lex.getRandomLexemeOfPOS(leaf.getPOS_ID());
      aSentence[j] = word;
    }
    sentenceV.addElement(aSentence);
  }
//----------------------------------------------------//
  private void writeToFile(Vector<String[]> sentenceV, int classID)
  {
    try
    {
      FileOutputStream out = new FileOutputStream("setn_class_" + classID + ".txt");
      
      String[] setn;
      String s;
      
      for (int i=0; i < sentenceV.size(); i++)
      {
        setn = sentenceV.elementAt(i);
        
        s = "@" + instCnt++ + ", ";
        out.write(s.getBytes());
        // wrt.print("@" + i + ", ");
        
        for (int j=0; j < setn.length; j++)
        {
          s = setn[j] + ", ";
          out.write(s.getBytes());
          // wrt.print(setn[j] + ", ");
        }
        
        s = "@" + classID + "\n";
        out.write(s.getBytes());
        // wrt.print("@" + classID + "\n");
      }
      
      out.flush();
      out.close();
    }
    catch (FileNotFoundException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    
  }
//----------------------------------------------------//
  /**
   * If 'instance' is :: { the, boy, who, likes, the, girl }
   * then there are 6 elements to be temporally presented, but
   * each sub-pattern (a word) are to be given in parallel 
   * as n-width bitpatterns.
   */  
  private final double runCompositeInput(Instance instance)
  {
    double error = 0.0;
    ExpandedPattern[] exPattern = instance.getExpandedNeuralNetData_NO_INFO();
    double[] pat;
    
    // for each word in sentence, designated by 'i'
    for (int i=0; i < exPattern.length; i++)
    {
      pat = exPattern[i].getPattern();
      
      // RAAM.printArray("pat0 pattern ", pat, 4);
      
      // set each bit in the word-pattern (width = # of words in dictionary)
      for (int j=0; j < pat.length; j++) inputLayer[j].setInput(pat[j]);
      
      // set feedback-input from the hidden units
      for (int j=0; j < nHiddenNodes; j++) inputLayer[j+realInputNodes].setInput(hiddenLayer[j].getInput() * 0.4);
      // for (int j=0; j < nHiddenNodes; j++) inputLayer[j+realInputNodes].setInput(hiddenLayer[j].nodeFunction());
      
      resetInput_HO();
      propagateAllInputs();
      error += backPropagate(instance, i);      
      updateWeights();
      // if (annCfg.updateWeightsPerPattern()) 
    }
    
    return error;
  }
//---------------------------------------------------------------------//
  /**
   * for test/evaluation after training epochs are complete!
   * no weightupdates or backprop
   */
  private Vector<double[]> runInput_final(Instance instance, boolean storeMetadata)
  {
    ExpandedPattern[] exPattern = instance.getExpandedNeuralNetData_NO_INFO();
    double[] pat;
    Vector<double[]> outputPrWord = new Vector<double[]>();

    // for each word in sentence, designated by 'i'
    for (int i=0; i < exPattern.length; i++)
    {
      pat = exPattern[i].getPattern();
      
      // set each bit in the word-pattern (width = # of words in dictionary)
      for (int j=0; j < pat.length; j++) inputLayer[j].setInput(pat[j]);
      
      // set feedback-input from the hidden units
      for (int j=0; j < nHiddenNodes; j++) inputLayer[j+realInputNodes].setInput(hiddenLayer[j].getInput() * 0.4);
      // for (int j=0; j < nHiddenNodes; j++) inputLayer[j+realInputNodes].setInput(hiddenLayer[j].nodeFunction());
      
      resetInput_HO();
      propagateAllInputs();
      
      if (storeMetadata) outputPrWord.addElement(getPatternFromLayer(nOutputNodes, outputLayer));
    }
    
    if (!storeMetadata) return null;
    
    
    /**
     * --- create/store the data for the FNN-MAPPER network!
     * As the SRN is supposed to have created a sufficiently structured temporal
     * representation of the entire sentence, we only need to store the ouput given
     * after the last word has been input.
     */ 
    double[] hiddenLayerPattern = getPatternFromLayer(hiddenLayer.length, hiddenLayer);
    Instance ins = new Instance(instance.getLabel(), null, instance.getClassID());
    ins.setPlainANN_data(hiddenLayerPattern);
    patternsFromSRN.addElement(ins);
    
    // RAAM.printArray("SRN hiddenlayer: ", hiddenLayerPattern, 4);
    // RAAM.printArray("SRN outputlayer: ", getPatternFromLayer(outputLayer.length, outputLayer), 4);
    
    return outputPrWord;
  }
//---------------------------------------------------------------------//
  private final void propagateAllInputs()
  {
    for (int i=0; i < nInputNodes; i++)  inputLayer[i].propagateForward();    
    for (int i=0; i < nHiddenNodes; i++) hiddenLayer[i].propagateForward();
  }
//-----------------------------------------------------------------//
  /**
   * for each input-step (a word) for current instance 'curPat',
   * we are backpropagating to learn to predict which word(s) may follow
   */
  protected double backPropagate(Instance curPat, int attribNo)
  {
    double wChange, errorK = 0, errorJ = 0, outputK, outputJ, patternError = 0.0;
    double d, momentumAdd;
    int j, i;
    
    computeInputAndHiddenLayerNFs();
    LegalSuccessor[] ls  = legalSuccessorV.elementAt(curPat.getClassID());
    boolean[] correctTbl = ls[attribNo].getLegalSuccessorTable();
    
    // for each outputNode to hiddenNodes
    for (i=0; i < outputLayer.length; i++)
    {
      if (correctTbl[i]) d = 1.0f; else d = 0.0;
      
      outputK       = outputLayer[i].nodeFunction();
      errorK        =  outputK * (1.0f - outputK) * (d - outputK);
      
      patternError += Math.abs((d - outputK));
      // patternError += ((d - outputK) * (d - outputK));
      
      outputLayer[i].tmpError = errorK;
      
      // for each backwardlink in outputNode[i]... update the linkWeight back to hiddenLayer-nodes
      for (j=0; j < outputLayer[i].backwardConnections.length; j++)
      {
        // wChange = (learningRate * errorK * hiddenLayer[j].nodeFunction()); // the pure weight change
        wChange = (learningRate * errorK * hiddenLayerNF[j]); // the pure weight change
        
        momentumAdd = (momentum * outputLayer[i].getBackwardPrevChangeAt(j));  // adding momentum-stuff! 
        outputLayer[i].backwardConnections[j].cumulateChanges(wChange + momentumAdd);
      }
    }
    
    // for each hiddenNode...to inputnodes
    for (i=0; i < hiddenLayer.length; i++)
    {
      outputJ = hiddenLayer[i].nodeFunction();
      errorJ  = (outputJ * (1 - outputJ)) * getSumOfWeightedErrors(hiddenLayer[i]);      
      
      for (j=0; j < hiddenLayer[i].backwardConnections.length; j++)
      {
        // wChange = (learningRate * (errorJ * inputLayer[j].nodeFunction())); // the pure weight change
        wChange = (learningRate * (errorJ * inputLayerNF[j])); // the pure weight change
        
        momentumAdd = (momentum * hiddenLayer[i].getBackwardPrevChangeAt(j));  // adding momentum-stuff!
        hiddenLayer[i].backwardConnections[j].cumulateChanges(wChange + momentumAdd);
      }
    }
    return patternError;
  }
//--------------------------------------------------------------------//
  protected void startTrainingX()
  {
    gui.clearOutput();
    rnd.setSeed(System.currentTimeMillis());

    gui.print("# of epochs : " + epochs + "\n");
    
    double epochError;
    int patternNo;
    Instance inst;

    patternsFromSRN              = new Vector<Instance>();
    Vector<Instance> tmpPatterns = new Vector<Instance>();
    Vector<Double> stats         = new Vector<Double>();

    long startS = System.currentTimeMillis();
    ///////////////////////////////////////////////////////////
    for (int i=0; i < epochs; i++)
    {
      gui.showProgress(i+1);
      
      ////////////////////////////////////
      /// ADD random pattern-order! HERE:
      for (int j=0; j < patterns_train.size(); j++) tmpPatterns.addElement(patterns_train.elementAt(j));
      epochError = 0;
      
      while (!tmpPatterns.isEmpty())
      {
        patternNo   = rnd.nextInt(tmpPatterns.size());
        inst        = tmpPatterns.elementAt(patternNo);        
        epochError += runCompositeInput(inst);
        
        tmpPatterns.remove(patternNo);
      }

      // usually a very, very bad idea...
      // if (annCfg.updateWeightsPerEpoch()) updateWeights();
      
      stats.addElement(new Double(epochError / patterns_train.size()));       
    }
    /////////////////////////////////////////////////////////     
    long endS = System.currentTimeMillis();
    
    printResultsX(((endS - startS) / 1000), stats, patterns_train);     
  }
//--------------------------------------------------------------------//
  /**
   * Finds the successor accuracy for a single word in a sentence
   */
  private double getSuccessorAccuracy(double slack, double[] output, int structID, int wordNo)
  {
    LegalSuccessor[] ls = legalSuccessorV.elementAt(structID);    
    boolean[] succTable = ls[wordNo].getLegalSuccessorTable();
    double okCnt = 0.0;

    //////------------ alternate way of determining degree of accuracy
    Vector<Integer> bitsON_correct = new Vector<Integer>();
    Vector<Integer> bitsON_actual;
    
    for (int i=0; i < succTable.length; i++)
    {
      if (succTable[i]) bitsON_correct.addElement(i);
    }
    
    bitsON_actual   = RAAM.findNHighestOuputs(bitsON_correct.size(), output);
    boolean allGood = RAAM.hasSameContent(bitsON_correct, bitsON_actual);
    
    if (allGood)  return 100.0;
    if (!allGood) return RAAM.sameContentRatio(bitsON_correct, bitsON_actual);
    
    //////--------------------------------------------------------------------
    
    
    //  find which bits (outputnodes) ACTUALLY are on
    for (int i=0; i < output.length; i++)
    {
      if (succTable[i]) // this output should be ON
      {
        if (output[i] >= (1 - slack)) okCnt++;
      }
      else // this output should be OFF
      {
        if (output[i] <= slack) okCnt++;
      }
    }
    
    double accuracyPct = (okCnt / output.length) * 100.0;
    // System.out.println("Getting accuracy for " + ls[wordNo].getPOSLabel() + " : " + accuracyPct);
    
    return accuracyPct;
  }
//--------------------------------------------------------------------//
  protected void printResultsX(long secs, Vector stats, Vector<Instance> patterns)
  {
    ///////////////////////////////////////////////////////////
    ///// run through all patterns again,  with no training
    boolean overallOK;
    double plainError, accuracy, totalAccuracy = 0.0, avgAcc = 0.0;
    Vector<double[]> outputPrWord;
    Instance instance;

    gui.clearOutput();
    // gui.print("FAILED PATTERN CLASSIFICATIONS:");
    
    for (int i=0; i < patterns.size(); i++)
    {
      totalAccuracy = 0.0;
      instance     = patterns.elementAt(i);      
      outputPrWord = runInput_final(instance, true);
      
      for (int j=0; j < outputPrWord.size(); j++)
      {
        accuracy = getSuccessorAccuracy(0.4, outputPrWord.elementAt(j), instance.getClassID(), j);
        totalAccuracy += accuracy;
        
        // RAAM.printArray("output for word #" + j, outputPrWord.elementAt(j), 4);
      }
      
      totalAccuracy /= outputPrWord.size();
      avgAcc += totalAccuracy;
      
      if (totalAccuracy >= 80) overallOK = true;
      else overallOK = false;
      
      
      if (!overallOK) gui.print("\n" + instance.getLabel() + " failed, accuracy = " + totalAccuracy);
      else gui.print("\n" + instance.getLabel() + " CORRECT, accuracy = " + totalAccuracy);
    }
    
    gui.print("\n------------------------------------");
    
    if (stats != null) 
    {
      plainError = ((Double)stats.lastElement()).doubleValue();
      gui.print("\n\nMSE for last epoch : " + plainError);
      gui.print("\nEpochs run         : " + epochs);
      gui.print("\nTime elapsed       : " + secs + " secs");
    }    
    gui.print("\nAVG. CLASSIFICATION RATE : " + avgAcc / patterns.size() + "%");
  }
//--------------------------------------------------------------------//
  public double[] testThisInput(Instance instance)
  {
    runInput_final(instance, false);
    return getPatternFromLayer(hiddenLayer.length, hiddenLayer);
  }
//--------------------------------------------------------------------//
}