package sharkey_mod_nlp.gui.tree;

import java.awt.event.*;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.image.*;
import java.awt.*;
import java.util.Vector;

import sharkey_mod_nlp.treemodel.*;

public class TreeCanvas extends Canvas implements MouseMotionListener, MouseListener
{
	private TData[] plyData;
	private Vector<Rectangle> nodeRects;
	private Dimension treeDimension;
	private boolean anyNodeSelected;
	
	private String[] txt;
	private int txtY;
	
	private Font myFontBIG;
	private Graphics2D g;
	private BufferStrategy buf;
	private Color nodeBackground;
	private ScrollPane jsp;
	private int scrollDistH, prevX, dragH_left, dragH_right, dragCnt;
	private int scrollDistV, prevY, dragV_left, dragV_right, dragV, dragH;
	
	public TreeCanvas(Spatializer spz, boolean centerNodes)
	{
		this.treeDimension   = spz.compileTree(centerNodes);
		this.plyData         = spz.getTreeData();
		this.nodeRects       = spz.getNodeRects();
		this.anyNodeSelected = false;
		
		nodeBackground = new Color(149, 226, 251);
		
		myFontBIG = new Font("Lucida Sans", Font.BOLD, 28);
		
		this.setBackground(Color.WHITE);
		addMouseMotionListener(this);
		this.addMouseListener(this);
	}
//----------------------------------------------------//
	public boolean isDoubleBuffered()
	{
		return true;
	}
//----------------------------------------------------//
	public void setScollPaneRef(ScrollPane sp)
	{
		jsp = sp;
	}
//----------------------------------------------------//
	private void initializeGraphics()
	{
		createBufferStrategy(2);
		buf = getBufferStrategy();
	}
//----------------------------------------------------//
	private boolean foundDragDirection(int x, int y)
	{
		if (dragCnt < 2) // find drag-direction
		{
			dragCnt++;
		
		       if (x < prevX) dragH_left++;
			else if (x > prevX) dragH_right++;
		  
		       if (y < prevY) dragV_left++;
		 	else if (y > prevY) dragV_right++;
		       
		       if (Math.abs(prevX - x) > Math.abs(prevY - y)) dragH++;
		  else if (Math.abs(prevY - y) > Math.abs(prevX - x)) dragV++; 
		  
			prevX = x;
			prevY = y;
			return false;
		}
		return true;
	}
//----------------------------------------------------//
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mouseClicked(MouseEvent e) {}	
	public void mouseReleased(MouseEvent e) 
	{ 
		setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR)); 
	}
//----------------------------------------------------//
	public void mousePressed(MouseEvent e)
	{
		dragCnt     = 0; 
		dragH_left  = 0;
		dragH_right = 0;
		dragV_left  = 0;
		dragV_right = 0;
		dragV       = 0;
		dragH       = 0;
		prevX       = e.getX();
		prevY       = e.getY();
		
		setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
	}
//----------------------------------------------------//
	public void mouseDragged(MouseEvent e)
	{
		if (!foundDragDirection(e.getX(), e.getY())) return;
		
		scrollDistH = 0;
		scrollDistV = 0;
		
		     if (dragH_left  > dragH_right) scrollDistH =  160;
		else if (dragH_right > dragH_left)  scrollDistH = -160;
		     
		     if (dragV_left  > dragV_right) scrollDistV =  64;
		else if (dragV_right > dragV_left)  scrollDistV = -64;
		
		int posH = jsp.getHAdjustable().getValue();
		int posV = jsp.getVAdjustable().getValue();

		     if (dragH >= dragV) scrollDistV = 0;
		else if (dragV >  dragH) scrollDistH = 0;
		
		jsp.setScrollPosition(posH + scrollDistH, posV + scrollDistV);
		
		dragCnt     = 0; 
		dragH_left  = 0;
		dragH_right = 0;
		dragV_left  = 0;
		dragV_right = 0;
	}
//----------------------------------------------------//
	public void mouseMoved(MouseEvent e)
	{
		if (g == null) return;
		
		int rX, rY, rW, rH;
		
		int x = e.getX();
		int y = e.getY();
		
		for (int i=0; i < nodeRects.size(); i++)
		{
			rX = (int)nodeRects.elementAt(i).getX();
			rY = (int)nodeRects.elementAt(i).getY();
			rW = (int)nodeRects.elementAt(i).getWidth();
			rH = (int)nodeRects.elementAt(i).getHeight();
			
			if (x >= rX && y >= rY && x <= (rX+rW) && y <= (rY+rH))
			{
				g.setColor(Color.YELLOW);
				g.drawRoundRect(rX, rY, rW, rH, 10, 10);
				g.drawRoundRect(rX+1, rY+1, rW-2, rH-2, 10, 10);
				
				g.setColor(new Color(184,207,232));
				g.fillRoundRect(rX+2, rY+2, rW-3, rH-3, 10, 10);
				
				g.setColor(Color.RED);
				repaintNodeContents(i+1);
				
				buf.show();
				anyNodeSelected = true;
				return;
			}
		}
		if (anyNodeSelected)
		{
			displayTree(true);
			anyNodeSelected = false;
		}
	}
//----------------------------------------------------//
	private void repaintNodeContents(int nodeNo)
	{
		IDisplayableNode[] nodeRow;
		IDisplayableNode n;
		
		for (int i=0; i < plyData.length; i++)
		{
      nodeRow = plyData[i].getNodeRow();
			
			for (int j=0; j < nodeRow.length; j++)
			{
				nodeNo--;
				n = nodeRow[j];
				if (nodeNo == 0)
				{
					txt = ((IGenericNode)n).getTextualValue();
					txtY = 14;
					
					for (int z=0; z < txt.length; z++)
					{
						g.drawString(txt[z], n.getX()+6, n.getY()+txtY);
						txtY += 12;
					}					
					return;
				}				
			}
		}
	}
//----------------------------------------------------//
	public void displayTree(boolean show)
	{
		if (buf == null)
		{
			initializeGraphics();
			if (buf == null) return;
		}
		
		g = (Graphics2D)buf.getDrawGraphics();		
		if (g == null) return;
		
		// g.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
    // g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		
		IDisplayableNode n, child;
		IDisplayableNode[] nodeRow;
		Vector<IGenericNode> children;
		
		//////////////// draw all links first //////////
		for (int i=0; i < plyData.length; i++)
		{
			nodeRow = plyData[i].getNodeRow();
			
			for (int j=0; j < nodeRow.length; j++)
			{
				n        = nodeRow[j];				
				children = ((IGenericNode)n).getChildren();
				g.setColor(Color.BLACK);
				
				/////////////////////////////////////////////
				//// paint the links from parent to children
				for (int k=0; k < children.size(); k++)
				{
					child = (IDisplayableNode) children.elementAt(k);					
					g.drawLine(n.getCenterX(), n.getCenterY(), child.getCenterX(), child.getCenterY());
				}
			}
		}
		
		/////////////////////////
		/// Now draw the nodes
		for (int i=0; i < plyData.length; i++)
		{
			nodeRow = plyData[i].getNodeRow();
			
			for (int j=0; j < nodeRow.length; j++)
			{
				n = nodeRow[j];				

				g.setColor(Color.LIGHT_GRAY);
				g.fillRoundRect(n.getX(), n.getY(), n.getWidth(), n.getHeight(), 10, 10);
				
				g.setColor(Color.BLUE);
				g.drawRoundRect(n.getX(), n.getY(), n.getWidth(), n.getHeight(), 10, 10);
				g.drawRoundRect(n.getX()+1, n.getY()+1, n.getWidth()-2, n.getHeight()-2, 10, 10);
				
				g.setColor(Color.RED);
				
				txt = ((IGenericNode)n).getTextualValue();
				txtY = 14;
				
				if (txt != null)
				{
					for (int z=0; z < txt.length; z++)
					{
						g.drawString(txt[z], n.getX()+6, n.getY()+txtY);
						txtY += 12;
					}
				}
			}
		}
		if (show) buf.show();		
	}
//----------------------------------------------------//
	public void paint(Graphics ggg)
	{
		displayTree(true);
	}
//----------------------------------------------------//
	public void update(Graphics ggg) {}
//----------------------------------------------------//
	public Dimension getMinimumSize()
	{
		return treeDimension;
	}
//----------------------------------------------------//
	public Dimension getPreferredSize()
	{
		return treeDimension;
	}
//----------------------------------------------------//
	public Dimension getMaximumSize()
	{
		return treeDimension;
	}
//----------------------------------------------------//
}