package sharkey_mod_nlp.treemodel;

import java.util.Vector;
import java.io.*;
//-----------------------------------------------------------------//


public class ParseNode implements IGenericNode, IDisplayableNode, Serializable
{
  /**
   * tmp hardcode hack-crap to know the position/type of the
   * reserved bits in a bitpattern..
   */
  public static int BITNO_TERMINAL = -1; // not set..  
  
  private ParseNode parent;
  private Vector<IGenericNode> children;
  
  private String nodeLabel;
  private int ply, intValue, x, y, structureID_ref;
  private double[] bitPattern, decodedBitPattern;
  private String[] txtValue;
  
  private static int maxPly = 0, WIDTH, HEIGHT;
  private static int TOTAL_NUMBER_OF_BITS_SET_IN_PATTERN = -1;
//-----------------------------------------------------------------//
  public ParseNode(ParseNode parentNode, String label)
  {
    nodeLabel = label;
    parent    = parentNode;
    children  = new  Vector<IGenericNode>();
  }
//-----------------------------------------------------------------//
  public static void setNumberOfBitsON(int nb)
  {
    TOTAL_NUMBER_OF_BITS_SET_IN_PATTERN = nb;
  }
//-----------------------------------------------------------------//
  public static int getNumberOfBitsON()
  {
    return TOTAL_NUMBER_OF_BITS_SET_IN_PATTERN;
  }
//-----------------------------------------------------------------//
  /**
   * Creates a new, independent and unlinked node, with no parent or children -
   * but with copies of all primitive fields.
   */
  public ParseNode getPrimitiveClone()
  {
    ParseNode pn = new ParseNode(null, new String(nodeLabel));
    
    pn.ply             = ply;
    pn.intValue        = intValue;
    pn.x               = x;
    pn.y               = y;
    pn.structureID_ref = structureID_ref;
    
    pn.bitPattern = new double[bitPattern.length];
    for (int i=0; i < bitPattern.length; i++) pn.bitPattern[i] = bitPattern[i];
    
    pn.txtValue = new String[txtValue.length];
    for (int i=0; i < txtValue.length; i++) pn.txtValue[i] = new String(txtValue[i]);
    
    return pn;
  }
//-----------------------------------------------------------------//
  /**
   * for the current application; this is a number from 0 .. 7
   * - being the id of which of the 8 syntactic structures in the dataset.
   */
  public void setSyntacticStructureID(int id)
  {
    structureID_ref = id;
  }
//-----------------------------------------------------------------//
  public int getSyntacticStructureID()
  {
    return structureID_ref;
  }
//-----------------------------------------------------------------//
  public void setBitpattern(double[] pattern)
  {
    bitPattern     = pattern;
    BITNO_TERMINAL = pattern.length-1; // assign to LAST bit in pattern
  }
//-----------------------------------------------------------------//
  /**
   * Convenience-method for storing the outputs of the output-nodes
   * in the corresponding partition, when the decoding....
   */
  public void setDecodedBitpattern(double[] pattern)
  {
    decodedBitPattern = pattern;
  }
//-----------------------------------------------------------------//
  public void invertBitPattern()
  {
    for (int i=0; i < bitPattern.length; i++)
    {
      bitPattern[i] = 1 - bitPattern[i];
    }
  }
//-----------------------------------------------------------------//
  public void setBitNo(int bitNo, double value)
  {
    bitPattern[bitNo] = value;
  }
//-----------------------------------------------------------------//
  public double[] getBitpattern()
  {
    return bitPattern;
  }
//-----------------------------------------------------------------//
  public double[] getDecodedBitpattern()
  {
    return decodedBitPattern;
  }
//-----------------------------------------------------------------//
  public void setPOS_ID(int id)
  {
    this.intValue = id;
  }
//-----------------------------------------------------------------//
  public int getPOS_ID()
  {
    return intValue;
  }
//-----------------------------------------------------------------//
  public void setIntegerValue(int value)      { intValue = value; }  
  public void setTextualValue(String[] value) { txtValue = value; }  
  public Object       getContent()            { return null;      }
  public int          getIntegerValue()       { return intValue;  }
  public String[]     getTextualValue()       { return txtValue;  }
//-----------------------------------------------------------------//
  public void setPly(int myPly)
  {
    ply = myPly;    
    if (ply > maxPly) maxPly = ply;
  }
//-----------------------------------------------------------------//
  public int getMaxPlyDepth()
  {
    return maxPly;
  }
//-----------------------------------------------------------------//
  public int getPly()
  {
    return ply;
  }
//-----------------------------------------------------------------//
  public boolean isRoot()
  {
    return (parent == null);
  }
//-----------------------------------------------------------------//
  public boolean isLeafNode()
  {
    return children.isEmpty();
  }
//-----------------------------------------------------------------//
  public String getNodeLabel()
  {
    return nodeLabel;
  }
//-----------------------------------------------------------------//
  public ParseNode getParent()
  {
    return parent;
  }
//-----------------------------------------------------------------//
  public void setParent(ParseNode newParent)
  {
    parent = newParent;
  }
//-----------------------------------------------------------------//
  public Vector<IGenericNode> getChildren()
  {
    return children;
  }
//-----------------------------------------------------------------//
  public void addChild(IGenericNode child)
  {
    if (child == null || child.equals(parent) || child.equals(this))
    {
      System.out.println("Illegal childnode attempted added!");
    }
    else
    {
      children.addElement(child);
    }
  }
//-----------------------------------------------------------------//
  /**
   * IDisplayableNode interface methods
   */
//-----------------------------------------------------------------//
  public  int getX() { return x; }
  public  int getY() { return y; }
  
  public  int getWidth()  { return WIDTH;  }
  public  int getHeight() { return HEIGHT; }
  
  public  int getCenterX() { return x + (WIDTH  / 2); }
  public  int getCenterY() { return y + (HEIGHT / 2); }
  
  public  int getParentCenterX() { return (parent.getX() + (WIDTH  / 2)); }
  public  int getParentCenterY() { return (parent.getY() + (HEIGHT / 2)); }
    
  public  void setX(int xx) { x = xx; }
  public  void setY(int yy) { y = yy; }
  public  void setSize(int w, int h) { WIDTH = w; HEIGHT = h; }
  
  public  void changeX(int dx) { x += dx; }
  public  void changeY(int dy) { y += dy; }
//-----------------------------------------------------------------//
  
//-----------------------------------------------------------------//
  
//-----------------------------------------------------------------//
  
//-----------------------------------------------------------------//
  
//-----------------------------------------------------------------//
}
