package sharkey_mod_nlp.treemodel;

import java.awt.*;
import java.awt.Rectangle;
import java.util.Vector;

public class Spatializer
{
	private Tree<IGenericNode> tree;
	private TData[] plyData;
	private String[] labels;
	private Vector<Rectangle> nodeRectV;
	
	private int defaultNodeSpacing, plySpacing, groupSpacing;  
	
	/**
	 * Builds a spatial (2D) representation of 
	 * a tree with root as param 'root'.
	 */
	public Spatializer(Tree<IGenericNode> t, int nodeWidth, int nodeHeight, int nSpc, int grpSpc)
	{
		tree               = t;
		groupSpacing       = grpSpc;
		defaultNodeSpacing = nSpc;
		plySpacing         = (int)(nodeHeight * 2.2) + 20;
		
		// set the global width and height for all nodes
		IDisplayableNode root = (IDisplayableNode) t.getRoot();
		root.setSize(nodeWidth, nodeHeight);
	}
//----------------------------------------------------//
	public Vector<Rectangle> getNodeRects()
	{
		return nodeRectV;
	}
//----------------------------------------------------//
	/**
	 * Sets the labels to use instead of the numberic value
	 * in a node - this method uses the values as indices
	 * into the stringTable, and assumes integer values,
	 * as well as an implicit coupling between a node's
	 * value and a string....
	 */
	public void setLabels(String[] labels)
	{
		this.labels = labels;
	}
//----------------------------------------------------//
	public TData[] getTreeData()
	{
		return plyData;
	}
//----------------------------------------------------//
	public Dimension compileTree(boolean postCenterNodes)
	{
		Vector<IGenericNode>[] plyRows = tree.getNodesAsPly_Arrays();
		
		IDisplayableNode root = (IDisplayableNode) tree.getRoot();
		root.setX(0);
		root.setY(0);
		
		plyData = new TData[plyRows.length];
		
		for (int i=0; i < plyData.length; i++) plyData[i] = new TData(plyRows[i], labels);
		
		int x, y, grpWidth;
		IDisplayableNode[] nodeGrp;
		Vector<IDisplayableNode[]> groupedNodes;
		
		////////////////////////////////////////////////////////////////
		////// Now set the inital coordinates, using default nodespacing
		y = root.getY() + plySpacing;
		/**
		 * iterate over all ROWS in the tree: except root, (ply 0)
		 * which has pre-set coords.
		 */	
		
		for (int i=1; i < plyData.length; i++)
		{
			groupedNodes = plyData[i].getNodesByGroup();
			
			// for each group in { this ply }
			for (int j=0; j < groupedNodes.size(); j++)
			{
				nodeGrp  = groupedNodes.elementAt(j);
				
				if (nodeGrp.length == 1)
				{
					grpWidth = nodeGrp[0].getWidth();
				}
				else
				{
					grpWidth = (nodeGrp.length * nodeGrp[0].getWidth()) + ((nodeGrp.length-1) * defaultNodeSpacing);
				}
				
				x = nodeGrp[0].getParentCenterX() - (grpWidth / 2);
				nodeGrp[0].setX(x);
				nodeGrp[0].setY(y);
				
        // for each node[k] in group[j]
				for (int k=1; k < nodeGrp.length; k++)
				{
					x += defaultNodeSpacing + nodeGrp[k-1].getWidth();
					
					nodeGrp[k].setX(x);					
					nodeGrp[k].setY(y);
				}
			}
			y += plySpacing;
		}
		
		//////////////////////////////////////////////////////////////
		/// look for nodes overlapping in the x-dimension, and fix it 
		for (int i=2; i < plyData.length; i++)
		{
			fixHorizIntersection(plyData[i]);
		}
		
    /////////////////////////////////////////////////////////
		/// now re-center parents relative to their child-groups
		/// direction = bottom-up :: (+ don't send in the root-layer).		
		if (postCenterNodes)
		{
			for (int i=plyData.length-1; i > 0; i--) reCenterNodes(plyData[i]);
		}		
		// else reCenterNodes(plyData[plyData.length-1]);
		
		Dimension dim = computeTreeBounds();
		
    /////////////////////////////////////
		/// Store the rects for each node
		nodeRectV = new Vector<Rectangle>();
		
		for (int i=0; i < plyData.length; i++)
		{
			groupedNodes = plyData[i].getNodesByGroup();
			
			// for each group in { this ply }
			for (int j=0; j < groupedNodes.size(); j++)
			{
				nodeGrp = groupedNodes.elementAt(j);
				
        // for each node[k] in group[j]
				for (int k=0; k < nodeGrp.length; k++)
				{
					nodeRectV.addElement(new Rectangle(nodeGrp[k].getX(), nodeGrp[k].getY(), nodeGrp[k].getWidth(), nodeGrp[k].getHeight()));
				}
			}
		}
		//---------------------------
		
		return dim;
	}
//----------------------------------------------------//
	private Dimension computeTreeBounds()
	{
		int minX, maxX=0, minY, maxY=0;
		
		IDisplayableNode[] row = plyData[0].getNodeRow();
		minX = row[0].getX();
		minY = row[0].getY();
		
		for (int i=0; i < plyData.length; i++)
		{
			row = plyData[i].getNodeRow();
			
			if (row.length < 1) continue;
			
			maxY = row[0].getY() + row[0].getHeight();
			
			if (row[0].getX() < minX)
			{
				minX = row[0].getX();
			}
			
			if ((row[row.length-1].getX() + row[row.length-1].getWidth()) > maxX)
			{
				maxX = row[row.length-1].getX() + row[row.length-1].getWidth();
			}
		}
		
		/////////////////////////////////////////
		// make the minX & minY -ZERO-, + border
		for (int i=0; i < plyData.length; i++)
		{
			row = plyData[i].getNodeRow();
			
			for (int j=0; j < row.length; j++)
			{
				row[j].changeX(-(minX-8));
				row[j].changeY(-(minY-8));
			}
		}
		
		Dimension dim = new Dimension((maxX - minX)+18, (maxY - minY)+18); 
		
		// we need to add some xtra pixels to avoid the scrollpane-bars... ;)
		return dim;
	}
//----------------------------------------------------//
	private void fixHorizIntersection(TData aRow)
	{
		IDisplayableNode[] siblingNodes1        = null;		
		IDisplayableNode[] siblingNodes2        = null;
		Vector<IDisplayableNode[]> groupedNodes = aRow.getNodesByGroup();
		
		int x1_end, x2_start, xTrans;
		
		// checking for space-clash between nodegroup 'i' and 'i+1'
		for (int i=0; i < groupedNodes.size()-1; i++)
		{
			siblingNodes1 = groupedNodes.elementAt(i);
			siblingNodes2 = groupedNodes.elementAt(i+1);
				
			x1_end   = siblingNodes1[siblingNodes1.length-1].getX() + siblingNodes1[0].getWidth();
			x2_start = siblingNodes2[0].getX();
				
			////////////////////////////////////////////
			///// Aha, two nodegroups are intersecting!
			if ((x1_end + groupSpacing) >= x2_start)
			{
				xTrans = ((x1_end + groupSpacing) - x2_start);

				// move all nodes in group[i] (xTrans/2) pixels to the LEFT					
				translateNodes(siblingNodes1, -(xTrans/2), 0);
					
         // move all nodes in group[i+1] (xTrans/2) pixels to the RIGHT					
				translateNodes(siblingNodes2, (xTrans/2), 0);

        // Also, add same LEFT movement to all PREVIOUS rects!
				// for each previous nodegroup :-: (will NOT enter loop on i=0)
				for (int j=i-1; j >= 0; j--)
				{
					translateNodes(groupedNodes.elementAt(j), -(xTrans/2), 0);
				}
			}
		}
	}
//----------------------------------------------------//
	/**
	 * THERE IS SOMETHING WRONG WITH THIS SHIT!!!
	 */
	private void reCenterNodes(TData aRow)
	{
		IDisplayableNode[] siblingNodes         = null;		
		Vector<IDisplayableNode[]> groupedNodes = aRow.getNodesByGroup();
		
		IDisplayableNode parent;
		int x, grpCenter, grpWidth;
		//////////////////////////////////////////////
		/// Center parents over their child-groups
		for (int i=0; i < groupedNodes.size(); i++)
		{
			siblingNodes = groupedNodes.elementAt(i);
			
			// any node in a group will do - we want the parent (of the group)
			parent = (IDisplayableNode) ((IGenericNode)siblingNodes[0]).getParent();

			grpWidth  = aRow.getGroupWidth(siblingNodes);
			grpCenter = siblingNodes[0].getX() + (grpWidth / 2);
			x         = (grpCenter - (parent.getWidth() / 2));
			
			parent.setX(x);
		}
	}
//----------------------------------------------------//
	private void translateNodes(IDisplayableNode[] siblingNodes, int dx, int dy)
	{
		for (int i=0; i < siblingNodes.length; i++)
		{
			siblingNodes[i].changeX(dx);
			siblingNodes[i].changeY(dy);
		}
	}
//----------------------------------------------------//
}