package sharkey_mod_nlp.treemodel;

import java.util.Vector;

/**
 * @author Sveinung
 */
public final class TData
{
	// this contains all nodes in one ply of our tree
	private IDisplayableNode[] nodeRow;
	
	public TData(Vector<IGenericNode> v, final String[] labels)
	{
		IGenericNode n;
		nodeRow = new IDisplayableNode[v.size()];
		
		// i iterates over ALL nodes in a ply
		for (int i=0; i < v.size(); i++)
		{
			n = v.elementAt(i);
			nodeRow[i] = (IDisplayableNode) n;
			
			if (labels == null)
			{
				
			}
			else if (n.getIntegerValue() < 0 || n.getIntegerValue() >= labels.length)
			{
				// n.setTextualValue(n.getIntegerValue() + "");
			}
			else n.setTextualValue(new String[] {labels[n.getIntegerValue()]} );
		}
	}
//------------------------------------------------------------------//
	/**
	 * returns the nodes in *this* ply, grouped by parents.
	 * This method is only used during initial computation
	 * of coordinates / spacings, and not for subsequent painting
	 * of the tree...  (so it doesn't matter if it is efficient)
	 */
	public Vector<IDisplayableNode[]> getNodesByGroup()
	{
		Vector<IDisplayableNode[]> nGrps = new Vector<IDisplayableNode[]>();
		
		IDisplayableNode[] grp = null;
		IDisplayableNode nDisp;
		IGenericNode nGen;
		int siblings, idx;
		
		for (int i=0; i < nodeRow.length;)
		{
			nDisp = nodeRow[i];
			nGen  = (IGenericNode) nDisp;
			 
			if (nGen.getParent() != null)
			{
				siblings = nGen.getParent().getChildren().size();
				
				grp = new IDisplayableNode[siblings];
				
				idx = 0;
				for (int j=i; j < (siblings + i); j++)
				{
					grp[idx++] = nodeRow[j];
				}
				i += siblings;
			}
			else // it's probably the root, create a "group" with 1 element 
			{
				grp = new IDisplayableNode[1];
				grp[0] = nodeRow[i];				
				i++;
			}
			nGrps.addElement(grp);
		}		
		return nGrps;
	}
//------------------------------------------------------------------//
	public IDisplayableNode[] getNodeRow()
	{
		return nodeRow;
	}
//------------------------------------------------------------------//
	/**
	 * This method only works correctly AFTER all nodes have been 
	 * given their initial coordinates: Because the 'nodeSpacing'
	 * variable will arbitrarlity change during configuration of
	 * the tree, that parameter cannot be used for computing the size
	 * of a group of nodes.
	 */
	public int getGroupWidth(IDisplayableNode[] nodeGrp)
	{
		if (nodeGrp.length == 1)
		{
			return nodeGrp[0].getWidth();
		}
		
		int end   = nodeGrp.length - 1;
		int width = (nodeGrp[end].getX() + nodeGrp[end].getWidth()) - nodeGrp[0].getX();
				
		return width;
	}
//------------------------------------------------------------------//
}