package wang.env;

import java.util.*;

import wang.graph.GraphData;
import wang.lang.*;
//-------------------------------------------------------------//
public class Environment
{
  private static Word[] codeWords;
  private static Meaning[] meanings;

  /**
   * number of correct interpretations of 
   * all meanings, per epoch.
   */
  private static int[] decode_stats;
  private static GraphData data;
  
  private Vector agents;
  private int epochs;
  
  public Environment()
  {
    agents = new Vector();
    epochs = 1;
  }
//-------------------------------------------------------------//
  public void updateStats(int curEpoch)
  {
    decode_stats[curEpoch]++;    
    data.setCoordinate(curEpoch, decode_stats[curEpoch], curEpoch);
  }
//-------------------------------------------------------------//
  public static final int[] getCommunicationStats()
  {
    return decode_stats;
  }
//-------------------------------------------------------------//
  public GraphData getGraphData()
  {
    return data;
  }
//-------------------------------------------------------------//
  /**
   * Enables external queries regarding the progress of the 
   * task which the agents currently working on...
   * Right now, this simply uses the global counter-variable
   * in the Agent-CLASS.
   */
  public int getProgress()
  {
    return Agent.getCurrentEpoch();
  }
//-------------------------------------------------------------//
  public final void setNumberOfEpochs(int n)
  {
    epochs = n;
  }
//-------------------------------------------------------------//
  public final int getNumberOfEpochs()
  {
    return epochs;
  }
//-------------------------------------------------------------//
  public final void setNumberOfMeanings(int n)
  {
    meanings  = new Meaning[n];    
    for (int i=0; i < n; i++) meanings[i] = new Meaning("M_" + (char)i, i);
  }
//-------------------------------------------------------------//
  public final void setNumberOfWords(int n)
  {
    codeWords = new Word[n];
    for (int i=0; i < n; i++) codeWords[i] = new Word("W_" + (char)i, i);
  }
//-------------------------------------------------------------//
  public final void activate()
  {
    data         = new GraphData(epochs);
    decode_stats = new int[epochs];
    for (int i=0; i < decode_stats.length; i++) decode_stats[i] = 0;
    
    Agent.setTurnbasedCommunication(true);    
    Agent ag;
    
    for (int i=0; i < agents.size(); i++)
    {
      ag = (Agent) agents.elementAt(i);
      ag.initCodecs(meanings.length, codeWords.length);
      ag.startAgent();
    }
  }
//-------------------------------------------------------------//
  @SuppressWarnings("unchecked")
  public synchronized void addAgent(Agent agent)
  {
    if (agents.isEmpty())
    {
      agents.addElement(agent);
    }
    else
    {
      if (!agents.contains(agent))
      {
        agents.addElement(agent);
      }
    }
  }
//-------------------------------------------------------------//
  private final Agent getAgent(int id)
  {
    Agent ag;
    
    for (int i=0; i < agents.size(); i++)
    {
      ag = (Agent) agents.elementAt(i);
      
      if (ag.getID() == id) return ag;
    }
    return null;
  }
//-------------------------------------------------------------//
  public final int getNumberOfMeanings()
  {
    return meanings.length;
  }
//-------------------------------------------------------------//
  public final int getNumberOfWords()
  {
    return codeWords.length;
  }
//-------------------------------------------------------------//
  /**
   * if param 'targetAgentID' == -1, message will be broadcast to 
   * ALL agents in the environment
   */
  public final synchronized boolean talkTo(Agent src, int dstAgentID, int cWord_id, int meaning_id)
  {
    Agent ag;
    
    /////////////////////////////////////////
    //// broadcast msg to all other agents
    if (dstAgentID == -1)
    {
      for (int i=0; i < agents.size(); i++)
      {
        ag = (Agent) agents.elementAt(i);
        if (ag.getID() != src.getID())
        {
          return ag.giveMessage(src, codeWords[cWord_id].getID(), meaning_id);
        }
      }
    }
    ///////////////////////////////////////////////////////
    //// send msg to agent whose id eqals 'targetAgentID'
    else
    {
      ag = getAgent(dstAgentID);
      
      if (ag != null)
      {
        return ag.giveMessage(src, codeWords[cWord_id].getID(), meaning_id);
      }
    }
    return false;
  }
//-------------------------------------------------------------//
}
