package wang.graph;

import java.awt.*;

import wang.gui.GraphCanvas;

public class Visualizer // implements Runnable
{
  public static int BORDER_X = 12;
  public static int BORDER_Y = 5;
  
  private static final int RES_X = 14;
  private static final int RES_Y = 3;
  
  private GraphData graphData;
  private Graphics2D g;
  private GraphCanvas canvas;
  private int[] xC, yC, xC1, yC1;
  private Color graphColor1, graphColor2, axisColor;
  
  // private Thread myThread;
  private boolean drawing;
//--------------------------------------------------------------//
  public Visualizer(GraphCanvas targetCanvas)
  {
    canvas    = targetCanvas;
    g         = null;
    
    graphColor1 = new Color(228,0,11);
    graphColor2 = new Color(249,99,120);
    axisColor   = new Color(20,152,23);
  }
//--------------------------------------------------------------//
  public void setDatasource(GraphData dataSrc)
  {
    this.graphData = dataSrc;
  }
//--------------------------------------------------------------//
  public void init()
  {
    canvas.init();
    
    /*
    if (myThread != null) 
    {
      drawing  = false;
      myThread = null;
      System.gc();
    }
    
    myThread = new Thread(this);
    myThread.start();
    */
  }
//--------------------------------------------------------------//
  public void stop()
  {
    this.drawing = false;
  }
//--------------------------------------------------------------//  
  public final void drawGraph(int yAxisMax)
  {
    if (g == null) 
    {
      g = (Graphics2D) canvas.getBufferGraphics();
      g.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
      g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
      g.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
      g.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
      g.setFont(new Font("Courier New", Font.PLAIN, 9));
    }
    
    xC = graphData.getData_x();
    yC = graphData.getData_y();
    
    xC1 = new int[xC.length];
    yC1 = new int[yC.length];

    int limX = canvas.getWidth(); //  - Visualizer.BORDER_X;
    int limY = canvas.getHeight() - Visualizer.BORDER_Y;
    
    // make the graph fill its given space...
    double scaleFx = ((double)limX / (double)xC.length);
    double scaleFy = ((double)limY / yAxisMax);
    
    //////////////////////////////////
    //// transform & scale 
    for (int i=1; i < xC.length; i++)
    {
      xC[i] = (int) (xC[i] * scaleFx);
      yC[i] = (int) (yC[i] * scaleFy);      
      xC[i-1] += Visualizer.BORDER_X;
    }

    for (int i=0; i < xC1.length; i++)
    {
      xC1[i] = xC[i]; 
      yC1[i] = yC[i]-1;
    }
    
    g.setColor(graphColor2);
    g.drawPolyline(xC1, yC1, xC1.length);
    g.setColor(graphColor1);
    g.drawPolyline(xC, yC, xC.length);
    
    drawAxis(yAxisMax);    
    canvas.flipBuffer();
  }
//--------------------------------------------------------------//
  private void drawAxis(int yAxisMax)
  {
    int mEpoch = xC.length / RES_X;
    int epCnt  = mEpoch;
    
    g.setColor(axisColor);
    
    g.drawLine(0, 0, 0, canvas.getHeight());
    g.drawLine(1, 0, 1, canvas.getHeight());    
    g.drawLine(0, canvas.getHeight()-1, canvas.getWidth(), canvas.getHeight()-1);
    g.drawLine(0, canvas.getHeight()-2, canvas.getWidth(), canvas.getHeight()-2);
    
    ////////////////////////////////////////////////
    //// compute & draw markers for x-axis
    int markSpace  = (canvas.getWidth() / RES_X);    
    int p          = 0;
    int fontMod    = 0;
    
    for (int i=0; i < RES_X; i++)
    {
      p += markSpace;
      g.setColor(new Color(20,152,23));
      g.drawLine(p, canvas.getHeight()-8, p, canvas.getHeight());
      
           if (epCnt < 10)                 fontMod = 2;  // 0..9
      else if (epCnt > 9 && epCnt  < 100)  fontMod = 4;  // 10..99
      else if (epCnt > 99 && epCnt < 1000) fontMod = 8;  // 100..999
      else if (epCnt > 999)                fontMod = 12; // 1000...9999
      
      g.setColor(new Color(131,132,191));
      g.drawString(""+epCnt, (p - fontMod), canvas.getHeight()-12);
      epCnt += mEpoch;
    }
    
    ////////////////////////////////////////////////
    //// compute & draw markers for y-axis
    markSpace = (canvas.getHeight() / RES_Y);    
    mEpoch    = yAxisMax / RES_Y;
    epCnt     = 0;
    
    p = 6;
    for (int i=0; i < RES_Y; i++)
    {
      g.drawLine(0, p, 8, p);
      g.drawString("" + (yAxisMax - epCnt), 10, p+4);
      
      epCnt += mEpoch;      
      p     += markSpace;      
    }
  }
//--------------------------------------------------------------//
  public void run()
  {
    drawing = true;
    
    while (drawing)
    {
      try
      {
        // drawGraph();        
        Thread.sleep(50);
      }
      catch (InterruptedException e) {}
    }    
    // myThread = null;
  }
}
