package wang.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

import wang.env.*;
import wang.graph.*;

public class Control extends JFrame implements ActionListener, ChangeListener
{
  private JButton runSim;
  private JSlider slider_meanings, slider_words, slider_epochs;
  private JTextField tf1, tf2, tf3, tfResults;
  
  private ProgressListener progressbar;
  private GraphCanvas vizCanvas;
  private Environment env;
//----------------------------------------------------------------//  
  public Control(Environment env)
  {
    JFrame.setDefaultLookAndFeelDecorated(true);
    this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    this.setBounds(300,50,600,550);
    this.setTitle("Language evolution simulation (Wang, 2005)  -  Simple version by Sveinung M");
    this.setResizable(false);
    this.env = env;
    
    progressbar = new ProgressListener(0, 50, this);
    progressbar.setTarget(env);
    
    runSim = new JButton("Run simulation");
    runSim.setActionCommand("run");
    runSim.addActionListener(this);
    
    slider_meanings = new JSlider(1, 100, 25);
    slider_words    = new JSlider(1, 100, 25);
    slider_epochs   = new JSlider(1, 2000, 100);
    
    slider_meanings.addChangeListener(this);
    slider_words.addChangeListener(this);
    slider_epochs.addChangeListener(this);
    
    tf1 = new JTextField(3);
    tf2 = new JTextField(3);
    tf3 = new JTextField(4);
    
    tf1.setEditable(false);
    tf2.setEditable(false);
    tf3.setEditable(false);
    
    tfResults = new JTextField(10);
    tfResults.setEditable(false);
    
    JerkPanel jp = new JerkPanel(306,410);
    jp.setBorder(new BevelBorder(BevelBorder.LOWERED));
    
    jp.addAt(new JLabel("Meanings"), 4, 6,   60, 25);
    jp.addAt(new JLabel("Words"),    4, 36,  60, 25);
    jp.addAt(new JLabel("Epochs"),   4, 66,  60, 25);
    
    jp.addAt(tf1, 60, 0+5+6,   40, 18);
    jp.addAt(tf2, 60, 30+5+6,  40, 18);
    jp.addAt(tf3, 60, 60+5+6,  40, 18);
    
    jp.addAt(slider_meanings, 100,    0+6, 200, 25);
    jp.addAt(slider_words,    100,   30+6, 200, 25);
    jp.addAt(slider_epochs,   100,   60+6, 200, 25);
    jp.addAt(runSim,            4,  120+6, 105, 25);
    jp.addAt(progressbar,      120, 120+6, 135, 25);
    
    JLabel resLabel = new JLabel("Degree of successful communication");
    jp.addAt(resLabel,    4, 380, 180, 30);
    jp.addAt(tfResults, 185, 385, 90, 18);
    
    
    //////////////////////////////////////
    //// The visualizer canvas component
    vizCanvas = new GraphCanvas(Color.WHITE, getWidth()-15);
    JPanel canvasPanel = new JPanel();
    canvasPanel.setBorder(new BevelBorder(BevelBorder.LOWERED));
    canvasPanel.add(vizCanvas);

    this.getContentPane().add(jp, BorderLayout.WEST);
    this.getContentPane().add(canvasPanel, BorderLayout.SOUTH);
    
    tf1.setText(slider_meanings.getValue() + "");
    tf2.setText(slider_words.getValue() + "");
    tf3.setText(slider_epochs.getValue() + "");
  }
//----------------------------------------------------------------------------//
  public void actionPerformed(ActionEvent e)
  {
    if (e.getActionCommand().equalsIgnoreCase(runSim.getActionCommand()))
    {
      runSim.setEnabled(false);
      
      int meanings = slider_meanings.getValue();
      int words    = slider_words.getValue();
      int epochs   = slider_epochs.getValue();
      
      env.setNumberOfMeanings(meanings);
      env.setNumberOfWords(words);
      env.setNumberOfEpochs(epochs);
      env.activate();
      
      progressbar.setMaximum(epochs);
      progressbar.startMe();
    }
  }
//----------------------------------------------------------------------------//  
  public void notifyCompletion()
  {
    runSim.setEnabled(true);
    
    int[] results = Environment.getCommunicationStats();
    float successRate = ((float)results[results.length-1] / (float)env.getNumberOfMeanings()) * 100.0f;
    tfResults.setText(successRate + "%");
    
    Visualizer viz = new Visualizer(vizCanvas);
    viz.setDatasource(env.getGraphData());
    viz.init();
    viz.drawGraph(env.getNumberOfMeanings());  
  }  
//----------------------------------------------------------------------------//
  public void stateChanged(ChangeEvent e)
  {
    if (e.getSource().equals(slider_meanings))
    {
      tf1.setText(slider_meanings.getValue() + "");
    }
    else if (e.getSource().equals(slider_words))
    {
      tf2.setText(slider_words.getValue() + "");
    } 
    else if (e.getSource().equals(slider_epochs))
    {
      tf3.setText(slider_epochs.getValue() + "");
    }
  }
//----------------------------------------------------------------------------//
}


/**
 * Self-contained ProgressBar component that listens to a source,
 * and updates itself accordingly. An internal thread takes care
 * of the 'listening'-part, and terminates when *finished*.
 */
final class ProgressListener extends JProgressBar implements Runnable
{
  private Environment env;
  private Thread myThread;
  private Control controller;
//--------------------------------------------------------//
  public ProgressListener(int min, int max, Control controller)
  {
    super(min, max);
    this.controller = controller;
  }
//--------------------------------------------------------//
  public void setTarget(Environment env)
  {
    this.env = env;
  }
//--------------------------------------------------------//
  public void startMe()
  {
    myThread = new Thread(this);
    myThread.start();
  }
//--------------------------------------------------------//  
  public void run()
  {
    while (getValue() < getMaximum())
    {
      try 
      {
        setValue(env.getProgress());
        Thread.sleep(100);
      }
      catch (InterruptedException e) {} 
    }    
    this.setValue(this.getMinimum());    
    this.myThread = null;
    
    controller.notifyCompletion();
  }
//--------------------------------------------------------//
}
