package sfm.gui;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemListener;
import java.util.Vector;
import javax.swing.*;

import sfm.dataset.Dataset;
import sfm.dataset.Instance;
import sfm.shp.*;
import sfm.shp.network.ANNConfig;
import sfm.shp.network.FNN;

public class GUI_CONCEPT2WORD extends GUI_SHELL implements ActionListener, ItemListener
{
  public GUI_CONCEPT2WORD()
  {
    annConfig = new ANNConfig(36,16,14);
    network   = new FNN(this, annConfig, "concept2word");
    
    createNetworkButt.setEnabled(false);
    createNetworkButt.addActionListener(this);
    trainButt.addActionListener(this);
    testButt.addActionListener(this);
    saveButt.addActionListener(this);
    loadButt.addActionListener(this);
    
    this.tf_Epochs.setText("250");    
    this.tf_inputNodes.setText(36 + "");
    this.tf_hiddenNodes.setText("16");
    this.tf_outputNodes.setText(14 + "");
    
    this.tf_Learningrate.setText("0.15");
    this.tf_margin.setText("0.1");
    this.tf_Momentum.setText("0.65");
    this.tf_Noise.setText("0.0");
    
    loadDataset.setEnabled(true);
    loadDataset.setVisible(true);
    loadDataset.addActionListener(this);
  }
//---------------------------------------------------------------------//
  public void actionPerformed(ActionEvent e)
  {
    String cmd = e.getActionCommand();
    
    // check for whatever special actions for this net....
    //... checks
    if (cmd.equalsIgnoreCase("load_dataset"))
    {
      loadDataset.setEnabled(false);
      addMapperDatasets();
      createNetworkButt.setEnabled(true);
    } 
    
    super.actionPerformed(e);
  }
//---------------------------------------------------------------------//
  private void addMapperDatasets()
  {
    FNN labelNet  = FNN.getNetwork("words");
    FNN shapeNet  = FNN.getNetwork("shapes");
    FNN lociNet_h = FNN.getNetwork("location_horizontal");
    FNN lociNet_v = FNN.getNetwork("location_vertical");
    
    double[] patInput, patTT;
    ShapeData shp;
    LabelData ld;
    RelLocDataH locH;
    RelLocDataV locV;

    int nShapes = shapeNet.getTrainingData().size();
    int nLoci   = lociNet_h.getTrainingData().size();
    nLoci       += lociNet_v.getTrainingData().size();
    
    Dataset ds1 = new Dataset("Concept-to-Label", nShapes + nLoci);
      
    ////////////////////////////////////////////
    /// Get the shape HL-activations & label TT
    for (int i=0; i < shapeV.size(); i++)
    {
      shp = shapeV.elementAt(i);
      ld  = getLabelByName(shp.getShapeName());
        
      shapeNet.runThisInputPattern(shp.getPattern(true));
      patInput = shapeNet.getLayerActivation(shapeNet.getHiddenLayer());
        
      labelNet.runThisInputPattern(ld.getPattern());
      patTT = labelNet.getLayerActivation(labelNet.getHiddenLayer());
        
      ds1.addInstance(new Instance(shp.getShapeName(), patInput, i, patTT));
    }
    
    ////////////////////////////////////////////////////////////
    /// Get the location (horizontal) HL-activations & label TT
    for (int i=0; i < lociH.size(); i++)
    {
      locH = lociH.elementAt(i);
      ld   = getLabelByName(locH.getPreposition_x1_to_x2());
        
      lociNet_h.runThisInputPattern(locH.getPattern());
      patInput = lociNet_h.getLayerActivation(lociNet_h.getHiddenLayer());
        
      labelNet.runThisInputPattern(ld.getPattern());
      patTT = labelNet.getLayerActivation(labelNet.getHiddenLayer());
        
      ds1.addInstance(new Instance(locH.getPreposition_x1_to_x2() + "/" + locH.getPreposition_x2_to_x1(), patInput, i, patTT));
    }
    
    
    ////////////////////////////////////////////////////////////
    /// Get the location (vertical) HL-activations & label TT
    for (int i=0; i < lociV.size(); i++)
    {
      locV = lociV.elementAt(i);
      ld   = getLabelByName(locV.getPreposition_y1_to_y2());
        
      lociNet_v.runThisInputPattern(locV.getPattern());
      patInput = lociNet_v.getLayerActivation(lociNet_v.getHiddenLayer());
        
      labelNet.runThisInputPattern(ld.getPattern());
      patTT = labelNet.getLayerActivation(labelNet.getHiddenLayer());
        
      ds1.addInstance(new Instance(locV.getPreposition_y1_to_y2() + "/" + locV.getPreposition_y2_to_y1(), patInput, i, patTT));
    }
    
    network.setDataset(ds1);
    System.out.println("dataset of conceptToWord: " + ds1.getNumberOfInstances());
  }
//---------------------------------------------------------------------//
}