package sfm.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;

import javax.swing.*;
import javax.swing.border.BevelBorder;

import nasseutils.NasseBox;

import sfm.dataset.Dataset;
import sfm.dataset.Instance;
import sfm.dataset.input.ReadFile;
import sfm.shp.*;
import sfm.shp.network.ANNConfig;
import sfm.shp.network.FNN;

public abstract class GUI_SHELL extends JPanel implements FilenameFilter, ActionListener, ItemListener
{
  protected static final String shapeDir = "datasets/shapes/";

  public static Vector<ShapeData> shapeV;
  public static Vector<LabelData> labelV;
  public static Vector<RelLocDataH> lociH;
  public static Vector<RelLocDataV> lociV;
  //-----------------------------------------------
  protected String suffix;
  protected String[] shapeFiles;
  
  protected JCheckBox cb_updating, cb_adaptiveLR;
  
  protected Vector<JTextField> annCfgV;
  protected JTextArea jtA_out;
  protected JTextField tf_Epochs, tf_inputNodes, tf_hiddenNodes, tf_outputNodes;
  protected JTextField tfTrainRatio, tfTrainSamps, tfTestSamps;
  protected JTextField tf_margin, tf_Learningrate, tf_Momentum, tf_Noise; 
  
  protected JPanel buttP;
  protected JProgressBar progressBar;
  protected JButton trainButt, testButt;
  protected JButton saveButt, loadButt, createNetworkButt;
  protected JButton loadDataset;
  
  protected ImagePanel imgP;
  protected boolean shpWordMapperDataAdded;
  
  protected FNN network;
  protected ANNConfig annConfig;
  
  private int epochsRunCurrentSession, progressCnt;

  public GUI_SHELL()
  {
    shpWordMapperDataAdded  = false;
    suffix                  = "BMP";
    epochsRunCurrentSession = 0;
    progressCnt             = 0;
    
    progressBar = new JProgressBar();
    progressBar.setDoubleBuffered(true);
    
    annCfgV = new Vector<JTextField>();
    
    //  ------- checkbox
    /*
    JCheckBox cb_force = new JCheckBox("Run until successful", false); 
    cb_force.addItemListener(this);
    cb_force.setName("forcerun");
    */
    
    cb_updating = new JCheckBox("Update weights per epoch", false);
    cb_updating.addItemListener(this);
    cb_updating.setName("update_pr_epoch");
    
    cb_adaptiveLR = new JCheckBox("Use adaptive Learning rate", true);
    cb_adaptiveLR.addItemListener(this);
    cb_adaptiveLR.setName("adaptiveLR");
    
    JPanel cb_updatingP = new JPanel(new GridLayout(2,1));
    cb_updatingP.add(cb_updating);
    cb_updatingP.add(cb_adaptiveLR);
    
    //------- checkbox
    
    createNetworkButt = new JButton("Create Network!");
    createNetworkButt.setActionCommand("makenetwork");
    
    trainButt = new JButton("Train");
    trainButt.setActionCommand("start_training");
    trainButt.setEnabled(false);
    
    testButt = new JButton("Test");
    testButt.setActionCommand("start_testing");    
    testButt.setEnabled(false);

    loadDataset = new JButton("Get dataset...");
    loadDataset.setActionCommand("load_dataset");
    loadDataset.setEnabled(false);
    loadDataset.setVisible(false);
    
    buttP = new JPanel();
    buttP.add(loadDataset);
    buttP.add(createNetworkButt);
    buttP.add(trainButt);
    buttP.add(testButt);
    
    JPanel trainButtP = new JPanel(new BorderLayout());
    JLabel dummyL3 = new JLabel("-"); dummyL3.setForeground(trainButtP.getBackground());
    
    trainButtP.add(buttP, BorderLayout.WEST);
    
    tf_Epochs       = new JTextField("100");
    tf_margin       = new JTextField("0.1");
    tf_Learningrate = new JTextField("0.25");
    tf_Momentum     = new JTextField("0.65");
    tf_Noise        = new JTextField("0.00");
    tf_Noise.setEditable(false);
    
    tf_inputNodes   = new JTextField("");
    tf_hiddenNodes  = new JTextField("");
    tf_outputNodes  = new JTextField("");
    
    // tf_inputNodes.setEditable(false);
    // tf_outputNodes.setEditable(false);
    tf_margin.setToolTipText("The strength of classification: its value is the lowest accepted output for each pattern...");
    
    tf_Learningrate.setBorder(new BevelBorder(BevelBorder.LOWERED));
    tf_margin.setBorder(new BevelBorder(BevelBorder.LOWERED));
    tf_Momentum.setBorder(new BevelBorder(BevelBorder.LOWERED));
    tf_Noise.setBorder(new BevelBorder(BevelBorder.LOWERED));
    tf_Epochs.setBorder(new BevelBorder(BevelBorder.LOWERED));
    
    JPanel cfgPanel = new JPanel(new GridLayout(11,2));
    cfgPanel.add(new JLabel("Margin"));           cfgPanel.add(tf_margin);
    cfgPanel.add(new JLabel("Epochs"));           cfgPanel.add(tf_Epochs);
    
    cfgPanel.add(new JLabel("Learning rate"));    cfgPanel.add(tf_Learningrate);
    cfgPanel.add(new JLabel("Momentum"));         cfgPanel.add(tf_Momentum);
    cfgPanel.add(new JLabel("Noise"));            cfgPanel.add(tf_Noise);    
    cfgPanel.add(new JLabel(""));                 cfgPanel.add(new JLabel(""));
    cfgPanel.add(new JSeparator());               cfgPanel.add(new JSeparator());
    
    cfgPanel.add(new JLabel("Input nodes   "));    cfgPanel.add(tf_inputNodes);
    cfgPanel.add(new JLabel("Hidden nodes  "));    cfgPanel.add(tf_hiddenNodes);
    cfgPanel.add(new JLabel("Output nodes  "));    cfgPanel.add(tf_outputNodes);
    
    cfgPanel.add(new JLabel(""));                 cfgPanel.add(new JLabel(""));
    // cfgPanel.add(new JLabel("Progress  "));       cfgPanel.add(progressBar);
    
    NasseBox nb = new NasseBox(null);
    nb.add(cfgPanel);

    JPanel wrapP = new JPanel(new BorderLayout(30,10));
    wrapP.add(cb_updatingP, BorderLayout.NORTH);
    wrapP.add(nb, BorderLayout.CENTER);    
    wrapP.add(trainButtP, BorderLayout.SOUTH);    
    // wrapP.add(loaderP, BorderLayout.EAST);
    
    /////////////////////////////////////
    ///// the output tray....
    jtA_out = new JTextArea(10,75);
    jtA_out.setBorder(new BevelBorder(BevelBorder.LOWERED));
    JScrollPane jsp = new JScrollPane(jtA_out, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
    
    JPanel progressP = new JPanel(new BorderLayout(4,15));
    progressP.add(progressBar, BorderLayout.WEST);
    progressP.add(new JSeparator(), BorderLayout.SOUTH);
    
    JPanel outputP = new JPanel(new BorderLayout(2,10));
    outputP.add(progressP, BorderLayout.NORTH);
    outputP.add(new JLabel("Actual classifications & MSE for dataset"), BorderLayout.WEST);
    outputP.add(jsp, BorderLayout.SOUTH);
    
    JPanel wrapCfg = new JPanel(new FlowLayout(FlowLayout.LEFT));
    wrapCfg.add(wrapP);
    
    JPanel bothP = new JPanel(new BorderLayout(10,16));
    bothP.add(wrapCfg, BorderLayout.WEST);
    bothP.add(outputP, BorderLayout.SOUTH);
    
    this.setLayout(new FlowLayout(FlowLayout.LEFT));
    this.add(bothP);
    
    saveButt = new JButton("Save Network");
    saveButt.setActionCommand("save");
    
    loadButt = new JButton("Load Network");
    loadButt.setActionCommand("getfromfile");
    
    this.add(saveButt);
    this.add(loadButt);
    
    annCfgV.addElement(tf_margin);
    annCfgV.addElement(tf_Epochs);
    annCfgV.addElement(tf_Learningrate);
    annCfgV.addElement(tf_Momentum);
    annCfgV.addElement(tf_Noise);
    annCfgV.addElement(tf_inputNodes);
    annCfgV.addElement(tf_hiddenNodes);
    annCfgV.addElement(tf_outputNodes);
    
    tf_margin.setName("margin");
    tf_Epochs.setName("epochs");
    tf_Learningrate.setName("learningrate");
    tf_Momentum.setName("momentum");
    tf_Noise.setName("noise");    
    tf_inputNodes.setName("inputnodes");
    tf_hiddenNodes.setName("hiddennodes");
    tf_outputNodes.setName("outputnodes");
    
    tf_inputNodes.setEditable(false);
    tf_outputNodes.setEditable(false);
  }  
//-----------------------------------------------------------------//
  public int getEpochsRunInCurrentSession()
  {
    return epochsRunCurrentSession; 
  }
//-----------------------------------------------------------------//
  public void actionPerformed(ActionEvent e)
  {
    String cmd = e.getActionCommand();

    if (cmd.equalsIgnoreCase("start_training"))
    {
      if (parseInputfields())
      {
        epochsRunCurrentSession += annConfig.getEpochs();
        
        this.createNetworkButt.setEnabled(false);
        this.testButt.setEnabled(false);
        this.trainButt.setEnabled(false);
        network.startTraining(annConfig);
      }
    }
    else if (cmd.equalsIgnoreCase("makenetwork"))
    {
      if (parseInputfields())
      {
        epochsRunCurrentSession = 0;
        
        this.trainButt.setEnabled(true);
        this.testButt.setEnabled(false);
        network.createNetwork(annConfig);
      }
    }
  }  
//-----------------------------------------------------------------//
  public void itemStateChanged(ItemEvent e)
  {
    if (e.getSource() instanceof JCheckBox)
    {
      JCheckBox cb = (JCheckBox) e.getSource();
      
      if (cb.getName().equalsIgnoreCase("forcerun"))
      {
        if (e.getStateChange() == ItemEvent.DESELECTED)
        {
          annConfig.setRunUntilSuccessful(false);
          tf_Epochs.setEnabled(true);
        }
        else
        {
          annConfig.setRunUntilSuccessful(true);
          tf_Epochs.setEnabled(false);
        }
      }
      else if (cb.getName().equalsIgnoreCase("update_pr_epoch"))
      {
        if (e.getStateChange() == ItemEvent.DESELECTED)
        {
          annConfig.setUpdateWeightsPerEpoch(false);
        }
        else
        {
          annConfig.setUpdateWeightsPerEpoch(true);
        }
      }
      else if (cb.getName().equalsIgnoreCase("adaptiveLR"))
      {
        if (e.getStateChange() == ItemEvent.DESELECTED)
        {
          annConfig.setUseAdaptiveLearningRate(false);
        }
        else
        {
          annConfig.setUseAdaptiveLearningRate(true);
        }
      }
    }      
  }  
//-----------------------------------------------------------------//
  public final void clearOutput()
  {
    jtA_out.setText("");
  }
//---------------------------------------------------------------------//
  public final void print(String s)
  {
    jtA_out.append(s);
  } 
//---------------------------------------------------------------------//
  public final void showProgress(int inc)
  {
    progressCnt += inc;
    progressBar.setValue(progressCnt);
    
    /*
    try
    {
      Thread.sleep(1);
    } catch (InterruptedException e) {}
    */
  }
//---------------------------------------------------------------------//  
  public final void notify_finished()
  {
    trainButt.setEnabled(true);
    createNetworkButt.setEnabled(true);
    testButt.setEnabled(true);
    
    progressBar.setValue(0);
    progressCnt = 0;
  }
//---------------------------------------------------------------------//
  public boolean accept(File dir, String name)
  {
    if (suffix.equals(""))
    {
      if (name.indexOf('.') == -1) return true;
      return false; 
    }
    
    if (name.toUpperCase().endsWith(suffix)) return true;
    return false;
  }
//---------------------------------------------------------------------//
  protected final void loadLabels()
  {
    final double posValue = 0.2;
    
    LabelData label1 = new LabelData("TRIANGLE");
    label1.setPOS(LabelData.POS_NOUN, posValue);
    
    LabelData label2 = new LabelData("RECTANGLE");
    label2.setPOS(LabelData.POS_NOUN, posValue);
    
    LabelData label3 = new LabelData("CIRCLE");
    label3.setPOS(LabelData.POS_NOUN, posValue);
    
    LabelData label4 = new LabelData("ABOVE");
    label4.setPOS(LabelData.POS_PREP, posValue);
    
    LabelData label5 = new LabelData("BELOW");
    label5.setPOS(LabelData.POS_PREP, posValue);
    
    LabelData label6 = new LabelData("LEFT");
    label6.setPOS(LabelData.POS_PREP, posValue);
    
    LabelData label7 = new LabelData("RIGHT");
    label7.setPOS(LabelData.POS_PREP, posValue);
    
    LabelData label8 = new LabelData("ON");
    label8.setPOS(LabelData.POS_PREP, posValue);

    labelV.addElement(label1);
    labelV.addElement(label2);
    labelV.addElement(label3);
    labelV.addElement(label4);
    labelV.addElement(label5);
    labelV.addElement(label6);
    labelV.addElement(label7);
    labelV.addElement(label8);
  }
//-----------------------------------------------------------------//
  protected final ShapeData getShapeByName(String name)
  {
    ShapeData shp;
    
    for (int i=0; i < shapeV.size(); i++)
    {
      shp = shapeV.elementAt(i);
      // System.out.println("shapeName : " + shp.getShapeName() + " :: " + name);
      if (shp.getShapeName().equalsIgnoreCase(name)) return shp;
    }
    return null;
  }
//---------------------------------------------------------------------//
  protected final LabelData getLabelByName(String name)
  {
    LabelData label;
    
    for (int i=0; i < labelV.size(); i++)
    {
      label = labelV.elementAt(i);
      if (label.getLabel().equalsIgnoreCase(name)) return label;
    }
    return null;
  }
//---------------------------------------------------------------------//
  protected final void loadShapes()
  {
    String sN;
    
    for (int i=0; i < shapeFiles.length; i++)
    {
      sN = shapeFiles[i].toUpperCase();
      if (sN.indexOf('.') != -1) sN = sN.substring(0, sN.indexOf('.'));
      
      ShapeData shp = ReadFile.readShapeData(shapeDir + shapeFiles[i], sN);
      shapeV.addElement(shp);
    }
  }
//---------------------------------------------------------------------//
  /**
   * Parses all textfields and sets the corresponding values in the config
   * template for our ANN; ANNConfig.
   * All flags from checkboxes are updated on-the-fly... 
   */
  protected final boolean parseInputfields()
  {
    JTextField tf = null;
    
    try
    {
      for (int i=0; i < annCfgV.size(); i++)
      {
        tf = annCfgV.elementAt(i);
        if (!tf.isEnabled()) continue;
        
        double value = Double.parseDouble(tf.getText().trim());
        
        boolean ok = annConfig.setValueFor(tf.getName(), value);        
        if (tf.isEnabled() && tf.isEditable())
        {  
          tf.setBackground(Color.WHITE);
        }
        
        // some buggy stuff; tried to set the value for a named target that were NOT found, stop.
        if (!ok) return false;
      }
    }
    catch (NumberFormatException e)
    {
      if (tf != null) tf.setBackground(Color.PINK);
      return false;
    }
    
    progressBar.setMinimum(0);
    progressBar.setMaximum(annConfig.getEpochs());
    
    return true;
  }
//---------------------------------------------------------------------//
 
  
  
//---------------------------------------------------------------------// 
  /**
   * non-critical Utility functions...
   */
  public static String formatNumber(double d, int fracDigits)
  {
    DecimalFormat decF = new DecimalFormat();
    DecimalFormatSymbols dfs = decF.getDecimalFormatSymbols();
    dfs.setDecimalSeparator('.');
    decF.setDecimalFormatSymbols(dfs);
    decF.setMaximumFractionDigits(fracDigits);
    
    return decF.format(d);
  }
//---------------------------------------------------------------------//
  /**
   * non-critical Utility function...
   */
  public static void printArray(String label, double[] arr, int fracDigits)
  {
    if (fracDigits == -1)
    {
      System.out.print(label + ": ");
      for (int i=0; i < arr.length; i++)
      {
        System.out.print(arr[i] + " ");
      }
      System.out.println("");
      return;
    }
    
    String formatted;
    String digs;
    int digCnt;
    
    System.out.print(label + ": ");
    for (int i=0; i < arr.length; i++) 
    {
      digs = "";
      
      formatted = formatNumber(arr[i], fracDigits);
      
      digCnt = formatted.substring(formatted.indexOf(".")+1, formatted.length()).length();
      if (digCnt < fracDigits)
      {
        for (int j=0; j < (fracDigits - digCnt); j++) digs += "0";
      }
      
      if (formatted.length() == 1)
      {
        digs = "";
        for (int j=0; j < fracDigits; j++) digs += "0";
        formatted += "." + digs;
      }
      else formatted += digs;
      
      if (!formatted.startsWith("-")) formatted = " " + formatted;
      
      System.out.print(formatted + " ");
    }
    System.out.println();
  }
  
  public static String _printArray(String label, double[] arr, int fracDigits)
  {
    String str = "";
    
    if (fracDigits == -1)
    {
      str+=(label + ": ");
      for (int i=0; i < arr.length; i++)
      {
        str += (arr[i] + " ");
      }
      str += ("\n");
      return str;
    }
    
    String formatted;
    String digs;
    int digCnt;
    
    str += (label + ": ");
    for (int i=0; i < arr.length; i++) 
    {
      digs = "";
      
      formatted = formatNumber(arr[i], fracDigits);
      
      digCnt = formatted.substring(formatted.indexOf(".")+1, formatted.length()).length();
      if (digCnt < fracDigits)
      {
        for (int j=0; j < (fracDigits - digCnt); j++) digs += "0";
      }
      
      if (formatted.length() == 1)
      {
        digs = "";
        for (int j=0; j < fracDigits; j++) digs += "0";
        formatted += "." + digs;
      }
      else formatted += digs;
      
      if (!formatted.startsWith("-")) formatted = " " + formatted;
      
      str+=(formatted + " ");
    }
    str += "\n";
    return str;
  }
//-----------------------------------------------------------------//
}





class ImagePanel extends JPanel
{
  private Vector<ShapeData> shpV;
  private Dimension myDim;
  private int shpPtr;
  
  public ImagePanel(Vector<ShapeData> shp, int w, int h)
  {
    shpV   = shp;
    myDim  = new Dimension(w, h);
    shpPtr = 0;
  }
//---------------------------------------------------------------------//
  public void setSelectedShape(int shpNo)
  {
    shpPtr = shpNo;
  }
//---------------------------------------------------------------------//
  public void paint(Graphics g)
  {
    super.paint(g);
    
    g.translate(16,16);
    g.setColor(Color.WHITE);
    
    Vector<PixelData> v = shpV.elementAt(shpPtr).getEdgePixels();    
    int x, y;
    
    for (int i=0; i < v.size(); i++)
    {
      x = v.elementAt(i).getX();
      y = v.elementAt(i).getY();
      g.drawLine(x, y, x, y);
    }
  }
//---------------------------------------------------------------------//  
  public void update(Graphics g)
  {
    // super.update(g);
  }
//---------------------------------------------------------------------//
  public Dimension getSize()          { return myDim; }
  public Dimension getMinimumSize()   { return myDim; }
  public Dimension getMaximumSize()   { return myDim; }
  public Dimension getPreferredSize() { return myDim; }
//---------------------------------------------------------------------//
}