package sfm.scene;

import java.util.*;

import sfm.dataset.*;
import sfm.gui.*;
import sfm.shp.*;
import sfm.shp.network.*;

public class SceneAnalyzer
{
  private Vector<Result> resultsV;
  
  private FNN labelNet, shapeNet, lociNet_h, lociNet_v;
  private Scene scene;
  
  public SceneAnalyzer(Scene scenex)
  {
    labelNet  = FNN.getNetwork("words");
    shapeNet  = FNN.getNetwork("shapes");
    lociNet_h = FNN.getNetwork("location_horizontal");
    lociNet_v = FNN.getNetwork("location_vertical");
    scene     = scenex;
    resultsV  = new Vector<Result>();
  }
//----------------------------------------------------------//
  public void analyzeScene()
  {
    ShapeData shp;
    
    for (int i=0; i < scene.getVerticalLocations(); i++)
    {
      for (int j=0; j < scene.getHorizontalLocations(); j++)
      {
        shp = scene.getSampleFrom(i, j);
        processVisualSample(shp, i, j);
      } 
    }
    
    /////////////////////////////////////////////////////////
    /// Now that all shapes have been found and interpreted,
    /// see if we can find the relative positioning...
    interpretLocations();
  }
//----------------------------------------------------------//
  private void processVisualSample(ShapeData shp, int row, int col)
  {
    shapeNet.runThisInputPattern(shp.getPattern(true));
    double[] shpActivation = shapeNet.getLayerActivation(shapeNet.getOutputLayer());
    int shapeID = getIndexOfHiValue(shpActivation);
    
    Vector<ShapeData> shapeRepository = ShapeData.getShapePool();
    Result result = new Result(shp, row, col);
    
    
    //////////////////////////////////////////////////////
    /// use the returned output-activation to find which
    /// shape it codes for.
    ShapeData storedShape;
    int tmpShpID;
    
    for (int i=0; i < shapeRepository.size(); i++)
    {
      storedShape = shapeRepository.elementAt(i);
      tmpShpID    = storedShape.getShapeID(); 
      
      if (shapeID == tmpShpID)
      {
        result.setInterpretedShapeID(shapeID);
      }
    }
    
    result.setXY(scene.getPositionOf(col), scene.getPositionOf(row));
    resultsV.addElement(result);
  }
//----------------------------------------------------------//
  private int getIndexOfHiValue(double[] pat)
  {
    int hiID   = 0;
    double max = pat[0];
    
    for (int i=1; i < pat.length; i++)
    {
      if (pat[i] > max)
      {
        max  = pat[i];
        hiID = i;
      }
    }
    
    return hiID;
  }
//----------------------------------------------------------//
  /**
   * For all the shapes found in the scene:
   * give the positions of two-and-two shapes,
   * all combinations, feed the horizontal and
   * vertical components (x and y) to the location nets. 
   */
  private void interpretLocations()
  {
    Result res1, res2;
    double[] patIn, patOut;
    
    for (int i=0; i < resultsV.size(); i++)
    {
      res1 = resultsV.elementAt(i);
      
      for (int j=0; j < resultsV.size(); j++)
      {
        if (j != i)
        {
          res2 = resultsV.elementAt(i);
          
          patIn = findHorizontalLocOf(res1.getX(), res2.getX());
          lociNet_h.runThisInputPattern(patIn);
          patOut = lociNet_h.getLayerActivation(lociNet_h.getOutputLayer());
          // decode ouputpattern... find horizontal prepsition concept
          int[] prepsH = interpretLocationActivation(patOut);
          // create new results of {shapename1,prep1,prep2,shapename2,prep2,prep1}
          
          patIn = findVerticalLocOf(res1.getY(), res2.getY());
          lociNet_v.runThisInputPattern(patIn);
          patOut = lociNet_h.getLayerActivation(lociNet_h.getOutputLayer());
          // decode ouputpattern... find vertical prepsition concept
          int[] prepsV = interpretLocationActivation(patOut);
          // create new results of {shapename1,prep1,prep2,shapename2,prep2,prep1}
        }
      }
    }
  }
//----------------------------------------------------------//
  private int[] interpretLocationActivation(double[] actPat)
  {
    double[] prep1 = new double[5];
    double[] prep2 = new double[5];
    
    int idx = 0;
    
    for (int i=0; i < prep1.length; i++) prep1[i] = actPat[idx++]; 
    for (int i=0; i < prep2.length; i++) prep2[i] = actPat[idx++];
    
    int prepID_1 = getIndexOfHiValue(prep1);
    int prepID_2 = getIndexOfHiValue(prep1);
    
    return new int[] {prepID_1, prepID_2};
  }
//----------------------------------------------------------//
  private double[] findHorizontalLocOf(int x1, int x2)
  {
    Vector<RelLocDataH> locPatterns_H  = GUI_SHELL.lociH;
    int x11, x22;
    
    for (int i=0; i < locPatterns_H.size(); i++)
    {
      x11 = locPatterns_H.elementAt(i).getX1();
      x22 = locPatterns_H.elementAt(i).getX2();
      
      if (x1 == x11 && x2 == x22)
      {
        return locPatterns_H.elementAt(i).getPattern();
      }
    }
    return null;
  }
//----------------------------------------------------------//
  private double[] findVerticalLocOf(int y1, int y2)
  {
    Vector<RelLocDataV> locPatterns_V  = GUI_SHELL.lociV;
    int y11, y22;
    
    for (int i=0; i < locPatterns_V.size(); i++)
    {
      y11 = locPatterns_V.elementAt(i).getY1();
      y22 = locPatterns_V.elementAt(i).getY2();
      
      if (y1 == y11 && y2 == y22)
      {
        return locPatterns_V.elementAt(i).getPattern();
      }
    }
    return null;
  }
//----------------------------------------------------------//
  
//----------------------------------------------------------//
}


//=============================================================//
final class Result
{
  private int x,y, row,col;
  private ShapeData shape;
  private int guessedShapeID;
  
  public Result(ShapeData shp, int roww, int coll)
  {
    shape = shp;
    row   = roww;
    col   = coll;
    
    guessedShapeID = -1;
  }
//----------------------------------------------------------//
  public void setXY(int xx, int yy)
  {
    x = xx;
    y = yy;
  }
//----------------------------------------------------------//
  public ShapeData getShape()
  {
   return shape; 
  }
//----------------------------------------------------------//
  public int getRow() { return row; }
  public int getCol() { return col; }
//----------------------------------------------------------//
  public int getY() { return y; }
  public int getX() { return x; }
//----------------------------------------------------------//
  public void setInterpretedShapeID(int id)
  {
    guessedShapeID = id;
  }
//----------------------------------------------------------//
  public int getInterpretedShapeID()
  {
    return guessedShapeID;
  }
//----------------------------------------------------------//
  
}