package sfm.shp;

import java.util.*;

public class LabelData
{
  private static int LABEL_CNT   = 0x00;
  private static int MAX_LETTERS = 0x00;
  
  private static final String[] letterTable_all =
  {
    "a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z"
  };
  
  private static final String[] letterTable_consonant =
  {
    "b","c","d","f","g","h","j","k","l","m","n","p","r","s","t","v","w","x","z"
  };
  
  // private static final String[] letterTable_vowel = { "a","e","i","o","q","u","y" };
  
  public static final int POS_NOUN = 0;
  public static final int POS_PRON = 1;
  public static final int POS_VERB = 2;
  public static final int POS_PREP = 3;
  public static final int POS_ADJ  = 4;  
  public static final int POS_ADV  = 5;  
  public static final int POS_QUE  = 6;
  
  public static final int POSINFO_BITSIZE = 7;
  
//--------------------------------------------------------------------------------//
  private String label;
  private String[] letters;
  private Vector<double[]> xpV;
  private double[] posInfo;
  private int labelID; // classID
  
  public LabelData(String str)
  {
    label   = str;
    letters = new String[str.length()];
    xpV     = null;
    posInfo = new double[POSINFO_BITSIZE];
    labelID = LabelData.LABEL_CNT++;
    
    for (int i=0; i < posInfo.length; i++) posInfo[i] = 0.0; // all POS-flags disabled by default 
    for (int i=0; i < str.length(); i++) letters[i] = str.charAt(i) + "";
    
    if (str.length() > MAX_LETTERS) MAX_LETTERS = str.length();
  }
//---------------------------------------------------------------------//
  public static int getExpandedPatternLength()
  {
    return POSINFO_BITSIZE + 2; // 2 = letter# + vowel/consonant info
  }
//---------------------------------------------------------------------//
  public void setPOS(int posID, double value)
  {
    posInfo[posID] = value; 
  }
//---------------------------------------------------------------------//
  public int getLabelID()
  {
    return labelID;
  }
//---------------------------------------------------------------------//
  public String getLabel() { return label; }  
//---------------------------------------------------------------------//
  private void createExpandedPattern()
  {
    xpV = new Vector<double[]>();
    
    for (int i=0; i < letters.length; i++)
    {
      double[] pattern = new double[2 + posInfo.length];
      
      pattern[0] = ((double)letterNo(letters[i]) / (double)letterTable_all.length);
      
      if (isConsonant(letters[i]))
      {
        pattern[1] = 0.0;
      }
      else
      {
        pattern[1] = 1.0;
      }
      
      xpV.addElement(pattern);
    }
    
    // set the POS part of each subpattern (POS-bits supplied with each subpat).
    for (int i=0; i < xpV.size(); i++)
    {
      double[] pattern = xpV.elementAt(i);
      for (int j=0; j < posInfo.length; j++) pattern[2+j] = posInfo[j];
    }
  }
//---------------------------------------------------------------------//
  public static int getPatternLength()
  {
    return POSINFO_BITSIZE + 2; 
  }
//---------------------------------------------------------------------//
  public double[] getPattern()
  {
    double[] pattern = new double[2+posInfo.length];
    
    pattern[0] = ((double)labelID / (double)LabelData.LABEL_CNT);
    pattern[1] = ((double)letters.length / (double) MAX_LETTERS);
 
    // add pos-info (7 bits)
    for (int i=0; i < posInfo.length; i++) pattern[2+i] = posInfo[i];
    
    return pattern;
  }
//------------------------------------------------------------------//
  public double[] getTrainingTarget()
  {
    double[] tt = new double[LABEL_CNT];
    
    for (int i=0; i < tt.length; i++)
    {
      tt[i] = 0.0;
    }
    
    tt[labelID] = 1.0;
    
    return tt;
  }
//---------------------------------------------------------------------//
  public Vector<double[]> getExpandedPattern()
  {
    if (xpV == null) createExpandedPattern();
    
    return xpV;
  }
//---------------------------------------------------------------------//
  private int letterNo(String c)
  {
    for (int i=0; i < letterTable_all.length; i++)
    {
      if (c.equalsIgnoreCase(letterTable_all[i])) return i;
    }
    
    System.out.println("letter " + c + " not found in lettertable!");
    return -1;
  }
//---------------------------------------------------------------------//
  private boolean isConsonant(String c)
  {
    for (int i=0; i < letterTable_consonant.length; i++)
    {
      if (c.equalsIgnoreCase(letterTable_consonant[i])) return true;
    }
    
    return false;
  }
//---------------------------------------------------------------------//
  
//---------------------------------------------------------------------//
}
