package sfm.shp.network;

import sfm.dataset.*;
import java.util.*;
import java.io.Serializable;

//------------------------------------------------------------------//
public class ANNConfig implements Serializable
{
  private Dataset dataset;
  private double learningR, momentum, noise, margin, trainRatio;
  private int epochs, preEpochs, inputNodes, hiddenNodes, outputNodes;
  private boolean runUntilSuccessful, updateWeightsPrEpoch, useNativeLib, usePrestructuring;
  private boolean useAdaptiveLearningRate;
  
  public ANNConfig(int inputs, int hidden, int outputs)
  {
    updateWeightsPrEpoch    = false; // ==> update for each pattern
    usePrestructuring       = true;
    useAdaptiveLearningRate = false;
    runUntilSuccessful      = false;
    useNativeLib            = false;
    
    inputNodes  = inputs;
    hiddenNodes = hidden;
    outputNodes = outputs;    
    learningR   = 0.20;
    momentum    = 0.65;
    noise       = 0.0;
    trainRatio  = 1.0;
    margin      = 0.1;
  }
//------------------------------------------------------------------//
  /**
   * Currently using only textfield components
   */
  public boolean setValueFor(String name, double value)
  {
    if (name.toLowerCase().startsWith("learn"))
    {
      this.learningR = value;
    }
    else if (name.toLowerCase().startsWith("moment"))
    {
      this.momentum = value;
    }
    else if (name.toLowerCase().startsWith("noise"))
    {
      this.noise = value;
    }
    else if (name.toLowerCase().startsWith("margin"))
    {
      this.margin = value;
    }
    else if (name.toLowerCase().startsWith("epoch"))
    {
      this.epochs = (int) value;
    }
    else if (name.toLowerCase().startsWith("pre_epoch"))
    {
      this.preEpochs = (int) value;
    }
    else if (name.toLowerCase().startsWith("input"))
    {
      this.inputNodes = (int) value;
    }
    else if (name.toLowerCase().startsWith("hidden"))
    {
      this.hiddenNodes = (int) value;
    }
    else if (name.toLowerCase().startsWith("output"))
    {
      this.outputNodes = (int) value;
    }
    else
    {
      System.out.println("WHAT THE..?!! ANNConfig.setValueFor(...):\ntargetname '" + name + "' not found!");
      return false;
    }
    return true;
  }
//------------------------------------------------------------------//  
  public double getTrainingRatio()
  {
    return trainRatio;
  }  
//------------------------------------------------------------------//  
  public void setTrainingRatio(double tRatio)
  {
    trainRatio = tRatio;
  }
//------------------------------------------------------------------//  
  public double getMargin()
  {
    return margin;
  }
//------------------------------------------------------------------//
  public void setMargin(double mrg)
  {
    margin = mrg;
  }
//------------------------------------------------------------------//  
  public int getEpochs()
  {
    return this.epochs;
  }
//------------------------------------------------------------------//  
  public void setEpochs(int epochx)
  {
    this.epochs = epochx;
  }
//------------------------------------------------------------------//  
  public int getPrestructuringEpochs()
  {
    return this.preEpochs;
  }
//------------------------------------------------------------------//  
  public void setPrestructuringEpochs(int epochx)
  {
    this.preEpochs = epochx;
  }
//------------------------------------------------------------------//
  public int getHiddenNodes()
  {
    return this.hiddenNodes;
  }
//------------------------------------------------------------------//
  public void setHiddenNodes(int hiddenNodes)
  {
    this.hiddenNodes = hiddenNodes;
  }
//------------------------------------------------------------------//
  public int getInputNodes()
  {
    return this.inputNodes;
  }
//------------------------------------------------------------------//
  public void setInputNodes(int inputNodes)
  {
    this.inputNodes = inputNodes;
  }
//------------------------------------------------------------------//
  public double getLearningR()
  {
    return this.learningR;
  }
//------------------------------------------------------------------//
  public void setLearningR(double learningR)
  {
    this.learningR = learningR;
  }
//------------------------------------------------------------------//
  public double getMomentum()
  {
    return this.momentum;
  }
//------------------------------------------------------------------//
  public void setMomentum(double momentum)
  {
    this.momentum = momentum;
  }
//------------------------------------------------------------------//
  public double getNoise()
  {
    return this.noise;
  }
//------------------------------------------------------------------//
  public void setNoise(double noise)
  {
    this.noise = noise;
  }
//------------------------------------------------------------------//
  public int getOutputNodes()
  {
    return this.outputNodes;
  }
//------------------------------------------------------------------//
  public void setOutputNodes(int outputNodes)
  {
    this.outputNodes = outputNodes;
  }
//------------------------------------------------------------------//
  public boolean isRunUntilSuccessful()
  {
    return this.runUntilSuccessful;
  }
//------------------------------------------------------------------//
  public void setRunUntilSuccessful(boolean runUntilSuccessful)
  {
    this.runUntilSuccessful = runUntilSuccessful;
  }
//------------------------------------------------------------------//
  public boolean updateWeightsPerEpoch()
  {
    return updateWeightsPrEpoch;
  }
//------------------------------------------------------------------//
  public boolean updateWeightsPerPattern()
  {
    return !updateWeightsPrEpoch;
  }
//------------------------------------------------------------------//
  public void setUpdateWeightsPerEpoch(boolean updateWeightsPrEpoch)
  {
    this.updateWeightsPrEpoch = updateWeightsPrEpoch;
  }
//------------------------------------------------------------------//
  public boolean isUseNativeLib()
  {
    return this.useNativeLib;
  }
//------------------------------------------------------------------//
  public void setUseNativeLib(boolean useNativeLib)
  {
    this.useNativeLib = useNativeLib;
  }
//------------------------------------------------------------------//
  public void setUsePrestructuring(boolean enabled)
  {
    usePrestructuring = enabled;
  }
//------------------------------------------------------------------//
  public boolean prestructuringEnabled()
  {
    return usePrestructuring;
  }
//------------------------------------------------------------------//
  public void setUseAdaptiveLearningRate(boolean flag)
  {
    this.useAdaptiveLearningRate = flag;
  }
//------------------------------------------------------------------//
  public boolean usingAdaptiveLearningRate()
  {
    return useAdaptiveLearningRate;
  }
//------------------------------------------------------------------//
  public Dataset getDataset()
  {
    return dataset;
  }
//------------------------------------------------------------------//
  public void setDataset(Dataset ds)
  {
    dataset = ds;
  }
//------------------------------------------------------------------//  
}